"""
Two Plus Four
=============

"""

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class TwoPlusFour(Cage):
    """
    Represents a capsule cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0, -1]),
        _NonLinearCageVertex(1, [0, 0, 1]),

        _LinearCageVertex(2, [2, 0, 0], False),
        _LinearCageVertex(3, [-2, 0, 0], False),
        _LinearCageVertex(4, [0, 2, 0], False),
        _LinearCageVertex(5, [0, -2, 0], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[2], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[2], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[3], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[3], _vertex_prototypes[1]),
        Edge(4, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(5, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(6, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(7, _vertex_prototypes[5], _vertex_prototypes[1])
    )

    _num_windows = 4
    _num_window_types = 1



"""
M3L3 Triangle
=============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex
from ...topology_graph import Edge
from ....reactions import GenericReactionFactory


class M3L3Triangle(Cage):
    """
    Represents a cage topology graph.

    Both `corner` and `linker` vertices require building blocks with
    two functional groups for this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | corners: (0, 1, 2)
        | linkers: (3, 4, 5)

    See :class:`.Cage` for more details and examples.

    """

    def __init__(
        self,
        corners,
        linkers,
        vertex_alignments=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
    ):
        """
        Initialize a :class:`.M3L3Triangle`.

        Parameters
        ----------
        corners : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all corner vertices on the topology
            graph.

        linkers : :class:`dict` or :class:`.BuildingBlock`
            Can be a :class:`dict` which maps the
            :class:`.BuildingBlock` instances to the ids of the
            vertices it should be placed on.

            Can also be a :class:`.BuildingBlock` instance, which
            should be placed on all linker vertices on the topology
            graph.

        vertex_alignments : :class:`dict`, optional
            A mapping from the id of a :class:`.Vertex`
            to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is referred to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        reaction_factory : :class:`.ReactionFactory`, optional
            The reaction factory to use for creating bonds between
            building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.


        """

        if isinstance(corners, dict):
            building_blocks = corners
        else:
            building_blocks = {corners: (0, 1, 2)}

        if isinstance(linkers, dict):
            linkers_dict = linkers
        else:
            linkers_dict = {linkers: (3, 4, 5)}

        building_blocks.update(
            (building_block, vertices)
            for building_block, vertices in linkers_dict.items()
        )

        super().__init__(
            building_blocks,
            vertex_alignments=vertex_alignments,
            reaction_factory=reaction_factory,
            num_processes=num_processes,
        )

    _x = 2*np.sqrt(3)/4
    _y = 2
    _vertex_prototypes = (
        _LinearCageVertex(0, [0, _x, 0]),
        _LinearCageVertex(1, [_y/2, -_x, 0]),
        _LinearCageVertex(2, [-_y/2, -_x, 0]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _LinearCageVertex.init_at_center(
            id=3,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCageVertex.init_at_center(
            id=4,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=5,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[0]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[3]),

        Edge(2, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[2], _vertex_prototypes[4]),

        Edge(4, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[0], _vertex_prototypes[5]),
    )

    _num_windows = 1
    _num_window_types = 1



"""
Kagome
======

"""

import numpy as np

from .cof import Cof
from .vertices import _LinearCofVertex, _NonLinearCofVertex
from ..topology_graph import Edge


class Kagome(Cof):
    """
    Represents a kagome COF topology graph.

    Building blocks with four and two functional groups are required
    for this topology graph.

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0.]),
        np.array([0., 0., 5/1.7321])
    )

    _vertex_prototypes = (
        _NonLinearCofVertex(0, (1/4)*_a + (3/4)*_b + (0.5)*_c),
        _NonLinearCofVertex(1, (3/4)*_a + (3/4)*_b + (1/2)*_c),
        _NonLinearCofVertex(2, (3/4)*_a + (1/4)*_b + (1/2)*_c),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCofVertex.init_at_center(
            id=3,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCofVertex.init_at_center(
            id=4,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCofVertex.init_at_center(
            id=5,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=6,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=7,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
            cell_shifts=((0, 0, 0), (-1, 1, 0)),
            lattice_constants=_lattice_constants
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=8,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
            cell_shifts=((0, 0, 0), (0, 1, 0)),
            lattice_constants=_lattice_constants
        ),

    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[3], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[3], _vertex_prototypes[1]),

        Edge(2, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[4], _vertex_prototypes[2]),

        Edge(4, _vertex_prototypes[5], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[5], _vertex_prototypes[2]),

        Edge(6, _vertex_prototypes[6], _vertex_prototypes[0]),
        Edge(
            id=7,
            vertex1=_vertex_prototypes[6],
            vertex2=_vertex_prototypes[1],
            periodicity=(-1, 0, 0),
        ),

        Edge(8, _vertex_prototypes[7], _vertex_prototypes[0]),
        Edge(
            id=9,
            vertex1=_vertex_prototypes[7],
            vertex2=_vertex_prototypes[2],
            periodicity=(-1, 1, 0),
        ),

        Edge(10, _vertex_prototypes[8], _vertex_prototypes[1]),
        Edge(
            id=11,
            vertex1=_vertex_prototypes[8],
            vertex2=_vertex_prototypes[2],
            periodicity=(0, 1, 0),
        ),
    )



def atom_to_json(atom):
    """
    Return a JSON representation of `atom`.

    Parameters
    ----------
    atom : :class:`.Atom`
        The atom to serialize.

    Returns
    -------
    :class:`dict`
        A JSON representation of `atom`.

    """

    return (
        atom.get_atomic_number(),
        atom.get_charge(),
    )


def bond_to_json(bond):
    """
    Return a JSON representation of `bond`.

    Parameters
    ----------
    bond : :class:`.Bond`
        The bond to serialize.

    Returns
    -------
    :class:`dict`
        A JSON representation of `bond`.

    """

    return (
        bond.get_atom1().get_id(),
        bond.get_atom2().get_id(),
        bond.get_order(),
        bond.get_periodicity(),
    )



"""
Terminal Alkyne Factory
=======================

"""
from .functional_group_factory import FunctionalGroupFactory
from .utilities import _get_atom_ids
from ..functional_groups import Alkyne


class TerminalAlkyneFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Alkyne` instances.

    Creates functional groups from substructures, which match the
    ``[*][C]#[C][H]`` functional group string.

    Examples
    --------
    You want to create a building block which has
    :class:`.Alkyne` functional groups, but only if they are terminal.
    You want the non-terminal carbon atom in those functional
    groups to be the *bonder* atom, and the terminal CH
    group to be the *deleter* atoms.

    .. code-block:: python

        import stk

        building_block = stk.BuildingBlock(
            smiles='C#CCCCCC#C',
            functional_groups=(stk.TerminalAlkyneFactory(), ),
        )

    You want to create a building block which has
    :class:`.Alkyne` functional groups. You want the carbon
    atoms to be the *bonder* atoms and you don't want any *deleter*
    atoms.

    .. code-block:: python

        import stk

        terminal_alkyne_factory = stk.TerminalAlkyneFactory(
            # The indices of the carbon atoms in the functional
            # group string (see docstring) are 1 and 2.
            bonders=(1, 2),
            deleters=(),
        )
        building_block = stk.BuildingBlock(
            smiles='C#CCCCCC#C',
            functional_groups=(terminal_alkyne_factory, ),
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1, ), deleters=(2, 3), placers=None):
        """
        Initialize a :class:`.TerminalAlkyneFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in _get_atom_ids('[*][C]#[C][H]', molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Alkyne(
                atom1=atoms[0],
                carbon1=atoms[1],
                carbon2=atoms[2],
                atom2=atoms[3],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



from .one_one_reaction import *  # noqa
from .one_two_reaction import *  # noqa
from .two_two_reaction import *  # noqa
from .reaction import Reaction  # noqa
from .ring_amine_reaction import *  # noqa
from .dative_reaction import *  # noqa



"""
Tournament
==========

"""

import numpy as np

from stk.molecular import Inchi
from .selector import Selector


class Tournament(Selector):
    """
    Yields batches of molecules through tournament selection.

    In tournament selection, a random number of batches is chosen from
    the population to undergo a competition. In each competition, the
    batch with the highest fitness value is yielded. This is repeated
    until `num_batches` are yielded.

    Examples
    --------
    *Yielding Single Molecule Batches*

    Yielding molecules one at a time. For example, if molecules need
    to be selected for mutation or the next generation.

    .. code-block:: python

        import stk

        # Make the selector.
        tournament = stk.Tournament(
            num_batches=5,
            batch_size=1
        )

        # Select the molecules.
        for selected, in tournament.select(population):
            # Do stuff with each selected molecule, like apply a
            # mutation to it to generate a mutant.
            mutation_record = mutator.mutate(selected)

    """

    def __init__(
        self,
        num_batches=None,
        batch_size=1,
        duplicate_molecules=True,
        duplicate_batches=True,
        key_maker=Inchi(),
        fitness_modifier=None,
        random_seed=None,
    ):
        """
        Initialize a :class:`.Tournament` instance.

        Parameters
        ----------
        num_batches : :class:`int`, optional
            The number of batches to yield. If ``None`` then yielding
            will continue forever or until the generator is exhausted,
            whichever comes first.

        batch_size : :class:`int`, optional
            The number of molecules yielded at once.

        duplicate_molecules : :class:`bool`, optional
            If ``True`` the same molecule can be yielded in more than
            one batch.

        duplicate_batches : :class:`bool`, optional
            If ``True`` the same batch can be yielded more than once.

        key_maker : :class:`.MoleculeKeyMaker`, optional
            Used to get the keys of molecules. If two molecules have
            the same key, they are considered duplicates.

        fitness_modifier : :class:`callable`, optional
            Takes the `population` on which :meth:`.select` is called
            and returns a :class:`dict`, which maps records in the
            `population` to the fitness values the :class:`.Selector`
            should use. If ``None``, the regular fitness values of the
            records are used.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        if fitness_modifier is None:
            fitness_modifier = self._get_fitness_values

        self._generator = np.random.RandomState(random_seed)
        if num_batches is None:
            num_batches = float('inf')

        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._batch_size = batch_size
        self._num_batches = num_batches
        super().__init__(
            key_maker=key_maker,
            fitness_modifier=fitness_modifier,
        )

    def _select_from_batches(self, batches, yielded_batches):
        # The tournament can only take place if there is more than 1
        # batch.
        while (
            len(batches) > 1
            and yielded_batches.get_num() < self._num_batches
        ):
            tournament_size = self._generator.randint(
                low=2,
                high=len(batches)+1
            )
            competitors = self._generator.choice(
                a=batches,
                size=tournament_size,
                replace=False
            )
            yield max(competitors)

            if not self._duplicate_molecules:
                batches = filter(
                    yielded_batches.has_no_yielded_molecules,
                    batches,
                )
            if not self._duplicate_batches:
                batches = filter(
                    yielded_batches.is_unyielded_batch,
                    batches,
                )
            if (
                not self._duplicate_molecules
                or not self._duplicate_batches
            ):
                batches = tuple(batches)



"""
Two Plus Three
==============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class TwoPlusThree(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0, 1]),
        _NonLinearCageVertex(1, [0, 0, -1]),

        _LinearCageVertex(2, [-1, -0.5*np.sqrt(3), 0], False),
        _LinearCageVertex(3, [1, -0.5*np.sqrt(3), 0], False),
        _LinearCageVertex(4, [0, 0.5*np.sqrt(3), 0], False),

    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[4]),
    )

    _num_windows = 3
    _num_window_types = 1



"""
Building Block
==============

"""


import logging
import os
import rdkit.Chem.AllChem as rdkit
from functools import partial
import warnings
import numpy as np

from ..functional_groups import FunctionalGroup
from ..atoms import Atom
from ..bonds import Bond
from .molecule import Molecule
from ...utilities import remake, flatten


logger = logging.getLogger(__name__)


class BuildingBlock(Molecule):
    """
    Represents a building block of a :class:`.ConstructedMolecule`.

    A :class:`BuildingBlock` can represent either an entire molecule or
    a molecular fragments used to construct a
    :class:`.ConstructedMolecule`. The building block uses
    :class:`.FunctionalGroup` instances to identify which atoms are
    modified during construction.

    """

    # Maps file extensions to functions which can be used to
    # create an rdkit molecule from that file type.
    _init_funcs = {
        '.mol': partial(
            rdkit.MolFromMolFile,
            sanitize=False,
            removeHs=False
        ),

        '.sdf': partial(
            rdkit.MolFromMolFile,
            sanitize=False,
            removeHs=False
        ),

        '.pdb': partial(
            rdkit.MolFromPDBFile,
            sanitize=False,
            removeHs=False,
            proximityBonding=False,
        ),
    }

    def __init__(
        self,
        smiles,
        functional_groups=(),
        placer_ids=None,
        random_seed=None,
        position_matrix=None,
    ):
        """
        Initialize a :class:`.BuildingBlock`.

        Notes
        -----
        The molecule is given 3D coordinates with
        :func:`rdkit.ETKDGv2`.

        Parameters
        ----------
        smiles : :class:`str`
            A SMILES string of the molecule.

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        random_seed : :class:`int`, optional
            Random seed passed to :func:`rdkit.ETKDGv2`. This
            argument is deprecated and will be removed in any
            version of :mod:`stk` released on, or after, 01/08/20.
            If you want to change the position matrix of the
            initialized building block, please use
            `position_matrix`.

        position_matrix : :class:`numpy.ndarray`, optional
            The position matrix the building block should use. If
            ``None``, :func:`rdkit.ETKDGv2` will be used to calculate
            it.

        Raises
        ------
        :class:`RuntimeError`
            If embedding the molecule fails.

        """

        if random_seed is not None:
            warnings.warn(
                'The BuildingBlock random seed parameter will be '
                'removed in any '
                'version of stk released on, or after, 01/08/20. '
                'If you want to change the position matrix of the '
                'initialized building block, please use the '
                'position_matrix parameter.',
                FutureWarning,
            )
        else:
            random_seed = 4

        molecule = rdkit.AddHs(rdkit.MolFromSmiles(smiles))
        if position_matrix is None:
            params = rdkit.ETKDGv2()
            params.randomSeed = random_seed
            if rdkit.EmbedMolecule(molecule, params) == -1:
                raise RuntimeError(
                    f'Embedding with seed value of {random_seed} '
                    'failed.'
                )
            rdkit.Kekulize(molecule)
        else:
            # Make sure the position matrix always holds floats.
            position_matrix = np.array(
                position_matrix,
                dtype=np.float64,
            )
            conformer = rdkit.Conformer(molecule.GetNumAtoms())
            for atom_id, position in enumerate(position_matrix):
                conformer.SetAtomPosition(atom_id, position)
            molecule.AddConformer(conformer)

        self._init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_molecule(
        cls,
        molecule,
        functional_groups=(),
        placer_ids=None,
    ):
        """
        Initialize from a :class:`.Molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to initialize from.

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        Returns
        -------
        :class:`.BuildingBlock`
             The building block. It will have the same atoms, bonds and
             atomic positions as `molecule`.

        """

        return cls.init(
            atoms=tuple(molecule.get_atoms()),
            bonds=tuple(molecule.get_bonds()),
            position_matrix=molecule.get_position_matrix(),
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_vabene_molecule(
        cls,
        molecule,
        functional_groups=(),
        placer_ids=None,
        random_seed=None,
        position_matrix=None,
    ):
        """
        Initialize from a :mod:`vabene.Molecule`.

        Notes
        -----
        The molecule is given 3D coordinates with
        :func:`rdkit.ETKDGv2()`.

        Parameters
        ----------
        molecule : :class:`vabene.Molecule`
            The molecule from which to initialize.

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        random_seed : :class:`int`, optional
            Random seed passed to :func:`rdkit.ETKDGv2`. This
            argument is deprecated and will be removed in any
            version of :mod:`stk` released on, or after, 01/08/20.
            If you want to change the position matrix of the
            initialized building block, please use
            `position_matrix`.

        position_matrix : :class:`numpy.ndarray`, optional
            The position matrix the building block should use. If
            ``None``, :func:`rdkit.ETKDGv2` will be used to calculate
            it.

        Returns
        -------
        :class:`.BuildingBlock`
             The building block.

        Raises
        ------
        :class:`RuntimeError`
            If embedding the molecule fails.

        """

        if random_seed is not None:
            warnings.warn(
                'The init_from_vabene_molecule random seed parameter '
                'will be removed in any '
                'version of stk released on, or after, 01/08/20. '
                'If you want to change the position matrix of the '
                'initialized building block, please use the '
                'position_matrix parameter.',
                FutureWarning,
            )
        else:
            random_seed = 4

        editable = rdkit.EditableMol(rdkit.Mol())
        for atom in molecule.get_atoms():
            rdkit_atom = rdkit.Atom(atom.get_atomic_number())
            rdkit_atom.SetFormalCharge(atom.get_charge())
            editable.AddAtom(rdkit_atom)

        for bond in molecule.get_bonds():
            editable.AddBond(
                beginAtomIdx=bond.get_atom1_id(),
                endAtomIdx=bond.get_atom2_id(),
                order=rdkit.BondType(bond.get_order()),
            )

        rdkit_molecule = editable.GetMol()
        rdkit.SanitizeMol(rdkit_molecule)
        rdkit_molecule = rdkit.AddHs(rdkit_molecule)

        if position_matrix is None:
            params = rdkit.ETKDGv2()
            params.randomSeed = random_seed
            if rdkit.EmbedMolecule(rdkit_molecule, params) == -1:
                raise RuntimeError(
                    f'Embedding with seed value of {random_seed} '
                    'failed.'
                )
        else:
            # Make sure the position matrix always holds floats.
            position_matrix = np.array(
                position_matrix,
                dtype=np.float64,
            )
            conformer = rdkit.Conformer(rdkit_molecule.GetNumAtoms())
            for atom_id, position in enumerate(position_matrix):
                conformer.SetAtomPosition(atom_id, position)
            rdkit_molecule.AddConformer(conformer)

        rdkit.Kekulize(rdkit_molecule)
        return cls.init_from_rdkit_mol(
            molecule=rdkit_molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init(
        cls,
        atoms,
        bonds,
        position_matrix,
        functional_groups=(),
        placer_ids=None,
    ):
        """
        Initialize a :class:`.BuildingBlock` from its components.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms of the building block.

        bonds : :class:`tuple` of :class:`.Bond`
            The bonds of the building block.

        position_matrix : :class:`numpy.ndarray`
            An ``(n, 3)`` position matrix of the building block.

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` holding the :class:`.FunctionalGroup`
            instances the building block should have, and / or
            :class:`.FunctionalGroupFactory` instances used for
            creating them.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        building_block = cls.__new__(cls)
        Molecule.__init__(
            self=building_block,
            atoms=atoms,
            bonds=bonds,
            position_matrix=position_matrix,
        )
        functional_groups = building_block._extract_functional_groups(
            functional_groups=functional_groups,
        )
        building_block._with_functional_groups(functional_groups)
        building_block._placer_ids = (
            building_block._normalize_placer_ids(
                placer_ids=placer_ids,
                functional_groups=building_block._functional_groups,
            )
        )
        return building_block

    @classmethod
    def init_from_file(
        cls,
        path,
        functional_groups=(),
        placer_ids=None,
    ):
        """
        Initialize from a file.

        Parameters
        ----------
        path : :class:`str`
            The path to a molecular structure file. Supported file
            types are:

                #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
                #. ``.pdb`` - PDB file

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        Raises
        ------
        :class:`ValueError`
            If the file type cannot be used for initialization.

        """

        _, extension = os.path.splitext(path)

        if extension not in cls._init_funcs:
            raise ValueError(
                f'Unable to initialize from "{extension}" files.'
            )
        # This remake needs to be here because molecules loaded
        # with rdkit often have issues, because rdkit tries to do
        # bits of structural analysis like stereocenters. remake
        # gets rid of all this problematic metadata.
        molecule = remake(cls._init_funcs[extension](path))

        return cls.init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )

    @classmethod
    def init_from_rdkit_mol(
        cls,
        molecule,
        functional_groups=(),
        placer_ids=None,
    ):
        """
        Initialize from an :mod:`rdkit` molecule.

        Parameters
        ----------
        molecule : :class:`rdkit.Mol`
            The molecule.

        functional_groups : :class:`iterable`, optional
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`, optional
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        Returns
        -------
        :class:`BuildingBlock`
            The molecule.

        """

        building_block = cls.__new__(cls)
        building_block._init_from_rdkit_mol(
            molecule=molecule,
            functional_groups=functional_groups,
            placer_ids=placer_ids,
        )
        return building_block

    def _init_from_rdkit_mol(
        self,
        molecule,
        functional_groups,
        placer_ids,
    ):
        """
        Initialize from an :mod:`rdkit` molecule.

        Parameters
        ----------
        molecule : :class:`rdkit.Mol`
            The molecule.

        functional_groups : :class:`iterable`
            An :class:`iterable` of :class:`.FunctionalGroup` or
            :class:`.FunctionalGroupFactory` or both.
            :class:`.FunctionalGroup` instances are added to the
            building block and :class:`.FunctionalGroupFactory`
            instances are used to create :class:`.FunctionalGroup`
            instances the building block should hold.
            :class:`.FunctionalGroup` instances are used to identify
            which atoms are modified during
            :class:`.ConstructedMolecule` construction.

        placer_ids : :class:`tuple` of :class:`int`
            The ids of *placer* atoms. These are the atoms which should
            be used for calculating the position of the building block.
            Depending on the values passed to `placer_ids`,
            and the functional groups in the building block, different
            *placer* ids will be used by the building block.

            #. `placer_ids` is passed to the initializer: the passed
               *placer* ids will be used by the building block.

            #. `placer_ids` is ``None`` and the building block has
               functional groups: The *placer* ids of the functional
               groups will be used as the *placer* ids of the building
               block.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. All atoms of the molecule will be used for
               *placer* ids.

        Returns
        -------
        None : :class:`NoneType`

        """

        atoms = tuple(
            Atom(a.GetIdx(), a.GetAtomicNum(), a.GetFormalCharge())
            for a in molecule.GetAtoms()
        )
        bonds = tuple(
            Bond(
                atom1=atoms[b.GetBeginAtomIdx()],
                atom2=atoms[b.GetEndAtomIdx()],
                order=(
                    9 if b.GetBondType() == rdkit.BondType.DATIVE
                    else b.GetBondTypeAsDouble()
                )
            )
            for b in molecule.GetBonds()
        )
        position_matrix = molecule.GetConformer().GetPositions()

        super().__init__(atoms, bonds, position_matrix)
        self._with_functional_groups(self._extract_functional_groups(
            functional_groups=functional_groups,
        ))
        self._placer_ids = self._normalize_placer_ids(
            placer_ids=placer_ids,
            functional_groups=self._functional_groups,
        )

    def _normalize_placer_ids(self, placer_ids, functional_groups):
        """
        Get the final *placer* ids.

        Parameters
        ----------
        placer_ids : :class:`tuple` of :class:`int`
            The ids of *placer* atoms or ``None``.

        functional_groups : :class:`iterable`
            The :class:`.FunctionalGroup` instances of the building
            block.

        Returns
        -------
        :class:`tuple` of :class:`int`
            Depending on the input  values, this function will return
            different things.

            #. `placer_ids` is a :class:`tuple` of :class`int`: the
                `placer_ids` will be returned.

            #. `placer_ids` is ``None`` and `functional_groups` is not
                empty: The *placer* ids of the functional groups will
                be returned.

            #. `placer_ids` is ``None`` and `functional_groups` is
               empty. The ids of all atoms in the building block will
               be returned.

        """

        if placer_ids is not None:
            return placer_ids

        if functional_groups:
            return tuple(flatten(
                functional_group.get_placer_ids()
                for functional_group in functional_groups
            ))

        return tuple(atom.get_id() for atom in self._atoms)

    def _extract_functional_groups(self, functional_groups):
        """
        Yield functional groups.

        The input can be a mixture of :class:`.FunctionalGroup` and
        :class:`.FunctionalGroupFactory`. The output yields
        :class:`.FunctionalGroup` instances only. Either those
        held directly in `functional_groups` or created by the
        factories in `functional_groups`.

        Parameters
        ----------
        functional_groups : :class:`iterable`
            Can be an :class:`iterable` of both
            :class:`.FunctionalGroup` and
            :class:`.FunctionalGroupFactory`.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group from `functional_groups`, or created
            by a factory in `functional_groups`.

        """

        for functional_group in functional_groups:
            if isinstance(functional_group, FunctionalGroup):
                yield functional_group
            else:
                # Else it's a factory.
                yield from functional_group.get_functional_groups(self)

    def _with_functional_groups(self, functional_groups):
        """
        Modify the molecule.

        """

        atom_map = {a.get_id(): a for a in self._atoms}
        self._functional_groups = tuple(
            fg.with_atoms(atom_map) for fg in functional_groups
        )
        return self

    def with_functional_groups(self, functional_groups):
        """
        Return a clone with specific functional groups.

        Parameters
        ----------
        functional_groups : :class:`iterable`
            :class:`.FunctionalGroup` instances which the clone
            should have.

        Returns
        -------
        :class:`.BuildingBlock`
            The clone. Has the same type as the original molecule.

        """

        return self.clone()._with_functional_groups(functional_groups)

    def _with_canonical_atom_ordering(self):
        ordering = rdkit.CanonicalRankAtoms(self.to_rdkit_mol())
        super()._with_canonical_atom_ordering()
        atom_map = {
            old_id: self._atoms[new_id]
            for old_id, new_id in enumerate(ordering)
        }
        self._functional_groups = tuple(
            functional_group.with_atoms(atom_map)
            for functional_group in self._functional_groups
        )
        self._placer_ids = tuple(
            atom_map[placer_id].get_id()
            for placer_id in self._placer_ids
        )
        return self

    def get_num_functional_groups(self):
        """
        Return the number of functional groups.

        Returns
        -------
        :class:`int`
            The number of functional groups in the building block.

        """

        return len(self._functional_groups)

    def get_functional_groups(self, fg_ids=None):
        """
        Yield the functional groups, ordered by id.

        Parameters
        ----------
        fg_ids : :class:`iterable` of :class:`int`, optional
            The ids of functional groups yielded. If ``None``, then
            all functional groups are yielded. Can be a single
            :class:`int`, if a single functional group is
            desired.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group of the building block.

        """

        if fg_ids is None:
            fg_ids = range(len(self._functional_groups))
        elif isinstance(fg_ids, int):
            fg_ids = (fg_ids, )

        for fg_id in fg_ids:
            yield self._functional_groups[fg_id]

    def clone(self):
        clone = super().clone()
        clone._functional_groups = self._functional_groups
        clone._placer_ids = self._placer_ids
        return clone

    def get_placer_ids(self):
        """
        Yield the ids of *placer* atoms.

        *Placer* atoms are those, which should be used to calculate
        the position of the building block.

        Yields
        ------
        :class:`int`
            The id of a *placer* atom.

        See Also
        --------
        :meth:`.FunctionalGroup.get_placer_ids`

        """

        yield from self._placer_ids

    def get_core_atom_ids(self):
        """
        Yield ids of atoms which form the core of the building block.

        This includes all atoms in the building block not part of a
        functional group, as well as any atoms in a functional group,
        specifically labelled as core atoms.

        Yields
        ------
        :class:`int`
            The id of a core atom.

        See Also
        --------
        :meth:`.FunctionalGroup.get_core_atom_ids`

        """

        functional_group_atom_ids = {
            atom_id
            for functional_group in self._functional_groups
            for atom_id in functional_group.get_atom_ids()
        }
        for atom in self._atoms:
            atom_id = atom.get_id()
            if atom_id not in functional_group_atom_ids:
                yield atom_id

        for functional_group in self._functional_groups:
            for atom_id in functional_group.get_core_atom_ids():
                yield atom_id

    def __str__(self):
        if self._functional_groups:
            fg_repr = f', {self._functional_groups!r}'
        else:
            fg_repr = ''

        smiles = rdkit.MolToSmiles(
            mol=rdkit.RemoveHs(self.to_rdkit_mol()),
        )
        return f'{self.__class__.__name__}({smiles!r}{fg_repr})'

    def __repr__(self):
        return str(self)



"""
Honeycomb
=========

"""

import numpy as np


from .cof import Cof
from .vertices import _LinearCofVertex, _NonLinearCofVertex
from ..topology_graph import Edge


class Honeycomb(Cof):
    """
    Represents a honeycomb COF topology graph.

    Building blocks with three and two functional groups are required
    for this topology graph.

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5/1.7321])
    )

    _vertex_prototypes = (
        _NonLinearCofVertex(0, (1/3)*_a + (1/3)*_b + (1/2)*_c),
        _NonLinearCofVertex(1, (2/3)*_a + (2/3)*_b + (1/2)*_c),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCofVertex.init_at_center(
            id=2,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=3,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=4,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        )
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[2], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[2], _vertex_prototypes[1]),

        Edge(2, _vertex_prototypes[3], _vertex_prototypes[0]),
        Edge(
            id=3,
            vertex1=_vertex_prototypes[3],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),

        Edge(4, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(
            id=5,
            vertex1=_vertex_prototypes[4],
            vertex2=_vertex_prototypes[1],
            periodicity=(-1, 0, 0),
        )
    )



"""
Random Building Block
=====================

"""

import numpy as np


from .mutator import MoleculeMutator
from ...records import MutationRecord
from ....molecule_records import MoleculeRecord


class RandomBuildingBlock(MoleculeMutator):
    """
    Substitutes random building blocks.

    This mutator takes a :class:`.ConstructedMolecule` and substitutes
    the building blocks with one chosen at random from a given set.

    Examples
    --------
    *Constructed Molecule Mutation*

    .. code-block:: python

        import stk

        # Create a molecule which is to be mutated.
        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        polymer = stk.MoleculeRecord(
            topology_graph=stk.polymer.Linear((bb1, bb2), 'AB', 3),
        )

        # Create molecules used to substitute building blocks.
        building_blocks = (
            stk.BuildingBlock(
                smiles='NC[Si]CCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            ),
            stk.BuildingBlock(
                smiles='NCCCCCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            ),
            stk.BuildingBlock(
                smiles='NC1CCCCC1N',
                functional_groups=[stk.PrimaryAminoFactory()],
            ),
        )

        # Create the mutator.

        def has_primary_amino_group(building_block):
            fg, = building_block.get_functional_groups(0)
            return fg.__class__ is stk.PrimaryAmino

        random_bb = stk.RandomBuildingBlock(
            building_blocks=building_blocks,
            is_replaceable=has_primary_amino_group,
        )

        # Mutate a molecule.
        mutation_record1 = random_bb.mutate(polymer)

        # Mutate the molecule a second time.
        mutation_record2 = random_bb.mutate(polymer)

    """

    def __init__(
        self,
        building_blocks,
        is_replaceable,
        name='RandomBuildingBlock',
        random_seed=None,
    ):
        """
        Initialize a :class:`.RandomBuildingBlock` instance.

        Parameters
        ----------
        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            A group of molecules which are used to replace building
            blocks in molecules being mutated.

        is_replaceable : :class:`callable`
            A function which takes a :class:`.BuildingBlock` and
            returns ``True`` or ``False``. This function is applied to
            every building block in the molecule being mutated.
            Building blocks which returned ``True`` are liable for
            substitution by one of the molecules in `building_blocks`.

        name : :class:`str`, optional
            A name to help identify the mutator instance.

        random_seed : :class:`bool`, optional
            The random seed to use.

        """

        self._building_blocks = building_blocks
        self._is_replaceable = is_replaceable
        self._name = name
        self._generator = np.random.RandomState(random_seed)

    def mutate(self, record):
        # Choose the building block which undergoes mutation.
        replaceable_building_blocks = tuple(filter(
            self._is_replaceable,
            record.get_molecule().get_building_blocks(),
        ))
        replaced_building_block = self._generator.choice(
            a=replaceable_building_blocks,
        )
        # Choose a replacement building block.
        replacement = self._generator.choice(self._building_blocks)

        # Build the new ConstructedMolecule.
        graph = record.get_topology_graph().with_building_blocks({
            replaced_building_block: replacement,
        })
        return MutationRecord(
            molecule_record=MoleculeRecord(graph),
            mutator_name=self._name,
        )



"""
Best
====

"""

import itertools as it

from stk.molecular import Inchi
from .selector import Selector


class Best(Selector):
    """
    Selects batches of molecules, highest fitness value first.

    Examples
    --------
    *Yielding Single Molecule Batches*

    Yielding molecules one at a time. For example, if molecules need
    to be selected for mutation or the next generation.

    .. code-block:: python

        import stk

        # Make the selector.
        best = stk.Best()

        # Select the molecules.
        for selected, in best.select(population):
            # Do stuff with each selected molecule, like apply a
            # mutation to it to generate a mutant.
            mutation_record = mutator.mutate(selected)

    *Yielding Batches Holding Multiple Molecules*

    Yielding multiple molecules at once. For example, if molecules need
    to be selected for crossover.

    .. code-block:: python

        import stk

        # Make the selector.
        best = stk.Best(batch_size=2)

        # Select the molecules.
        for selected in best.select(population):
            crossover_records = tuple(crosser.cross(selected))

    """

    def __init__(
        self,
        num_batches=None,
        batch_size=1,
        duplicate_molecules=True,
        duplicate_batches=True,
        key_maker=Inchi(),
        fitness_modifier=None,
    ):
        """
        Initialize a :class:`.Best` instance.

        Parameters
        ----------
        num_batches : :class:`int`, optional
            The number of batches to yield. If ``None`` then yielding
            will continue forever or until the generator is exhausted,
            whichever comes first.

        batch_size : :class:`int`, optional
            The number of molecule records yielded at once.

        duplicate_molecules : :class:`bool`, optional
            If ``True`` the same molecule can be yielded in more than
            one batch.

        duplicate_batches : :class:`bool`, optional
            If ``True`` the same batch can be yielded more than once.
            Duplicate batches can occur if the same molecule is found
            multiple times in a population.

        key_maker : :class:`.MoleculeKeyMaker`, optional
            Used to get the keys of molecules. If two molecules have
            the same key, they are considered duplicates.

        fitness_modifier : :class:`callable`, optional
            Takes the `population` on which :meth:`.select` is called
            and returns a :class:`dict`, which maps records in the
            `population` to the fitness values the :class:`.Selector`
            should use. If ``None``, the regular fitness values of the
            records are used.

        """

        if fitness_modifier is None:
            fitness_modifier = self._get_fitness_values

        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = num_batches
        self._batch_size = batch_size
        super().__init__(
            key_maker=key_maker,
            fitness_modifier=fitness_modifier,
        )

    def _select_from_batches(self, batches, yielded_batches):
        batches = sorted(batches, reverse=True)

        if not self._duplicate_molecules:
            batches = filter(
                yielded_batches.has_no_yielded_molecules,
                batches,
            )

        if not self._duplicate_batches:
            batches = filter(
                yielded_batches.is_unyielded_batch,
                batches,
            )

        yield from it.islice(batches, self._num_batches)



"""
Terminal Alkene Factory
=======================

"""

from .functional_group_factory import FunctionalGroupFactory
from .utilities import _get_atom_ids
from ..functional_groups import Alkene


class TerminalAlkeneFactory(FunctionalGroupFactory):
    # This docstring is a literal string, to silence a flake8 warning
    # about CH\ :sub:`2`. It's taking the backslash out of context.
    r"""
    Creates :class:`.Alkene` instances.

    Creates functional groups from substructures, which match the
    ``[*][C]([*])=[C]([H])[H]`` functional group string.

    Examples
    --------
    You want to create a building block which has
    :class:`.Alkene` functional groups, but only if they are terminal.
    You want the non-terminal carbon atom in those functional
    groups to be the *bonder* atom, and the terminal CH\ :sub:`2`
    group to be the *deleter* atoms.

    .. code-block:: python

        import stk

        building_block = stk.BuildingBlock(
            smiles='C=CCCCCC=C',
            functional_groups=(stk.TerminalAlkeneFactory(), ),
        )

    You want to create a building block which has
    :class:`.Alkene` functional groups, but only if they are terminal.
    You want the carbon atoms to be the *bonder* atoms and you don't
    want any *deleter* atoms.

    .. code-block:: python

        import stk

        terminal_alkene_factory = stk.TerminalAlkeneFactory(
            # The indices of the carbon atoms in the functional
            # group string (see docstring) are 1 and 3.
            bonders=(1, 3),
            deleters=(),
        )
        building_block = stk.BuildingBlock(
            smiles='C=CCCCCC=C',
            functional_groups=(terminal_alkene_factory, ),
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(
        self,
        bonders=(1, ),
        deleters=(3, 4, 5),
        placers=None,
    ):
        """
        Initialize a :class:`.TerminalAlkeneFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        ids = _get_atom_ids('[*][C]([*])=[C]([H])[H]', molecule)
        for atom_ids in ids:
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Alkene(
                carbon1=atoms[1],
                atom1=atoms[0],
                atom2=atoms[2],
                carbon2=atoms[3],
                atom3=atoms[4],
                atom4=atoms[5],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
Above Average
=============

"""

import numpy as np
import itertools as it

from stk.molecular import Inchi
from .selector import Selector


class AboveAverage(Selector):
    """
    Yields above average batches of molecules.

    Examples
    --------
    *Yielding Single Molecule Batches*

    Yielding molecules one at a time. For example, if molecules need
    to be selected for mutation or the next generation

    .. code-block:: python

        import stk

        # Make the selector.
        above_avg = stk.AboveAverage()

        # Select the molecules.
        for selected, in above_avg.select(population):
            # Do stuff with each selected molecule, like apply a
            # mutation to it to generate a mutant.
            mutation_record = mutator.mutate(selected)

    *Yielding Batches Holding Multiple Molecules*

    Yielding multiple molecules at once. For example, if molecules need
    to be selected for crossover.

    .. code-block:: python

        import stk

        # Make the selector.
        above_avg = stk.AboveAverage(batch_size=2)

        # Select the molecules.
        for selected in above_avg.select(population):
            # selected holds 2 molecules.
            crossover_records = tuple(crosser.cross(selected))

    """

    def __init__(
        self,
        num_batches=None,
        batch_size=1,
        duplicate_molecules=True,
        duplicate_batches=True,
        key_maker=Inchi(),
        fitness_modifier=None,
    ):
        """
        Initialize an :class:`.AboveAverage` instance.

        Parameters
        ----------
        num_batches : :class:`int`, optional
            The number of batches to yield. If ``None`` then yielding
            will continue forever or until the generator is exhausted,
            whichever comes first.

        batch_size : :class:`int`, optional
            The number of molecule records in each yielded
            :class:`.Batch`.

        duplicate_molecules : :class:`bool`, optional
            If ``True`` the same molecule can be yielded in more
            than one batch.

        duplicate_batches : :class:`bool`, optional
            If ``True`` the same batch can be yielded more than once.

        key_maker : :class:`.MoleculeKeyMaker`, optional
            Used to get the keys of molecules. If two molecules have
            the same key, they are considered duplicates.

        fitness_modifier : :class:`callable`, optional
            Takes the `population` on which :meth:`.select` is called
            and returns a :class:`dict`, which maps records in the
            `population` to the fitness values the :class:`.Selector`
            should use. If ``None``, the regular fitness values of the
            records are used.

        """

        if fitness_modifier is None:
            fitness_modifier = self._get_fitness_values

        self._duplicate_molecules = duplicate_molecules
        self._duplicate_batches = duplicate_batches
        self._num_batches = num_batches
        self._batch_size = batch_size
        super().__init__(
            key_maker=key_maker,
            fitness_modifier=fitness_modifier,
        )

    def _select_from_batches(self, batches, yielded_batches):
        mean = np.mean([
            batch.get_fitness_value() for batch in batches
        ])
        # Yield highest fitness batches first.
        batches = sorted(batches, reverse=True)
        # Yield only batches with a fitness larger than the mean.
        batches = it.takewhile(
            lambda batch: batch.get_fitness_value() > mean,
            batches
        )
        # Yield batches which are multiple times better than the mean
        # multiple times.
        batches = (
            batch
            for batch in batches
            for i in range(self._get_num_duplicates(batch, mean))
        )
        # If duplicate molecules are not allowed, allow only
        # batches with no yielded molecules.
        if not self._duplicate_molecules:
            batches = filter(
                yielded_batches.has_no_yielded_molecules,
                batches,
            )
        # If duplicate batches are not allowed, allow only
        # unyielded batches.
        if not self._duplicate_batches:
            batches = filter(
                yielded_batches.is_unyielded_batch,
                batches,
            )
        # Limit the number of yielded batches to _num_batches.
        yield from it.islice(batches, self._num_batches)

    def _get_num_duplicates(self, batch, mean):
        if self._duplicate_batches and self._duplicate_molecules:
            return int(batch.get_fitness_value() // mean)
        return 1



"""
Bidentate Square Planar
=======================

"""

from ..metal_complex import MetalComplex
from ..vertices import _MetalVertex, _BiDentateLigandVertex
from ...topology_graph import Edge


class BidentateSquarePlanar(MetalComplex):
    """
    Represents a square planar metal complex topology graph.

    Metal building blocks with at least four functional groups are
    required for this topology graph.

    Ligand building blocks with two functional groups are required
    for this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, )
        | ligands: (0, 1)

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (
        _MetalVertex(0, [0, 0, 0]),
    )
    _ligand_vertex_prototypes = (
        _BiDentateLigandVertex(1, [2.5, 2.5, 0]),
        _BiDentateLigandVertex(2, [-2.5, -2.5, 0]),
    )

    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[2.5, 0, 0],
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0, 2.5, 0],
        ),

        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
            position=[-1, 0, 0],
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
            position=[0, -1, 0],
        ),
    )



"""
Host Guest Complex
==================

"""

from .vertices import _HostVertex, _GuestVertex
from ..topology_graph import TopologyGraph


class Complex(TopologyGraph):
    """
    Represents a host-guest complex topology graph.

    Examples
    --------
    *Construction*

    You can use :class:`.ConstructedMolecule` instances as the host,
    but you should turn them into a :class:`.BuildingBlock` first

    .. code-block:: python

        import stk

        host = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusSix(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                    stk.BuildingBlock(
                        smiles='BrCC(Br)CBr',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
            ),
        )
        complex1 = stk.ConstructedMolecule(
            topology_graph=stk.host_guest.Complex(
                host=stk.BuildingBlock.init_from_molecule(host),
                guest=stk.BuildingBlock('[Br][Br]'),
            ),
        )

    *Changing the Position of the Guest*

    You can change the position and orientation of the guest, as well
    as its displacement

    .. code-block:: python

        guest = stk.BuildingBlock('[Br][Br]')
        complex2 = stk.ConstructedMolecule(
            topology_graph=stk.host_guest.Complex(
                host=stk.BuildingBlock.init_from_molecule(host),
                guest=guest,
                # Apply a rotation onto the guest molecule such that
                # the vector returned by get_direction() has the same
                # direction as [1, 1, 1].
                guest_start=guest.get_direction(),
                guest_target=[1, 1, 1],
                # Change the displacement of the guest.
                displacement=[5.3, 2.1, 7.1],
            ),
        )

    """

    def __init__(
        self,
        host,
        guest,
        guest_start=None,
        guest_target=None,
        displacement=(0, 0, 0),
        num_processes=1,
    ):
        """
        Initialize an instance of :class:`.Complex`.

        Parameters
        ----------
        host : :class:`.BuildingBlock`
            The host molecule.

        guest : :class:`.BuildingBlock`
            The guest molecule.

        guest_start : :class:`tuple` of :class:`float`, optional
            A direction vector which gets aligned with `guest_target`.

        guest_target : :class:`tuple` of :class:`float`, optional
            A direction vector which determines the rotation applied to
            the `guest`. A rotation such that `guest_start` is
            transformed into `guest_target` is applied.

        displacement : :class:`tuple` of :class:`float`, optional
            The translational offset of the guest.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        Raises
        ------
        :class:`TypeError`
            If `guest_start` or `guest_target` is defined but the other
            is not.

        """

        num_nones = sum(
            1 for vector in (guest_start, guest_target)
            if vector is None
        )
        if num_nones == 1:
            raise TypeError(
                'If guest_start or guest_target is defined, '
                'the other must be too.'
            )

        if guest_start is None:
            start = target = (1., 0., 0.)
        else:
            start = guest_start = tuple(guest_start)
            target = guest_target = tuple(guest_target)

        # Save the values as None, for __repr__.
        self._guest_start = guest_start
        self._guest_target = guest_target
        self._displacement = displacement

        super().__init__(
            building_block_vertices={
                host: (_HostVertex(0, [0, 0, 0]), ),
                guest: (
                    _GuestVertex(1, displacement, start, target),
                ),
            },
            edges=(),
            reaction_factory=None,
            construction_stages=(),
            num_processes=num_processes,
            edge_groups=(),
        )

    def clone(self):
        clone = super().clone()
        clone._guest_start = self._guest_start
        clone._guest_target = self._guest_target
        clone._displacement = self._displacement
        return clone

    def _run_reactions(self, state):
        return state

    def _get_scale(self, building_block_vertices):
        return 1

    def __repr__(self):
        return (
            f'host_guest.Complex('
            f'guest_start={self._guest_start!r}, '
            f'guest_target={self._guest_target!r}, '
            f'displacement={self._displacement!r})'
        )



"""
Progress Plotter
================

"""

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


plt.switch_backend('agg')


class ProgressPlotter:
    """
    Plots how a property changes during an EA run.

    The produced plot will show the EA generations on the x axis and
    the min, mean and max values of an attribute on the y axis.

    Examples
    --------
    *Plotting How Fitness Value Changes Across Generations*

    .. code-block:: python

        import stk

        # Initialize an EA somehow.
        ea = stk.EvolutionaryAlgorithm(...)

        generations = []
        for generation in ea.get_generations(10):
            generations.append(generation)

        # Make the plotter which plots the fitness change across
        # generations.
        progress = stk.ProgressPlotter(
            generations=generations,
            get_property=lambda record: record.get_fitness_value(),
            y_label='Fitness'
        )
        progress.write('fitness_plot.png')

    *Plotting How a Molecular Property Changes Across Generations*

    As an example, plotting how the number of atoms changes across
    generations

    .. code-block:: python

        # Initialize an EA somehow.
        ea = stk.EvolutionaryAlgorithm(...)

        generations = []
        for generation in ea.get_generations(10):
            generations.append(generation)

        # Make the plotter which plots the number of atoms across
        # generations.
        progress = stk.ProgressPlotter(
            generations=generations,
            get_property=lambda record:
                record.get_molecule().get_num_atoms(),
            y_label='Number of Atoms'
        )
        progress.write('fitness_plot.png')

    *Excluding Molecules From the Plot*

    Sometimes, you want to ignore some molecules from the plot you
    make. For example, If the fitness calculation failed on a
    molecule, you not want to include in a plot of fitness.

    .. code-block:: python

        import stk

        # Initialize an EA somehow.
        ea = stk.EvolutionaryAlgorithm(...)

        generations = []
        for generation in ea.get_generations(10):
            generations.append(generation)

        # Make the plotter which plots the fitness change across
        # generations.
        progress = stk.ProgressPlotter(
            generations=generations,
            get_property=lambda record: record.get_fitness_value(),
            y_label='Fitness'
            # Only plot records whose unnormalized fitness value is not
            # None, which means the fitness calculation did not fail.
            filter=lambda record:
                record.get_fitness_value(normalized=False) is not None,
        )
        progress.write('fitness_plot.png')

    """

    def __init__(
        self,
        generations,
        get_property,
        y_label,
        filter=lambda record: True,
    ):
        """
        Initialize a :class:`ProgressPlotter` instance.

        Parameters
        ----------
        generations : :class:`iterable` of :class:`.Generation`
            The generations of the EA, which are plotted.

        get_property : :class:`callable`
            A :class:`callable` which takes a :class:`.MoleculeRecord`
            and returns a property value of that molecule, which is
            used for the plot. The :class:`callable` must return a
            valid value for each
            :class:`.MoleculeRecord` in `generations`.

        y_label : :class:`str`
            The y label for the produced graph.

        filter : :class:`callable`, optional
            Takes an :class:`.MoleculeRecord` and returns
            ``True`` or ``False``. Only records which return ``True``
            are included in the plot. By default, all records will be
            plotted.

        """

        self._get_property = get_property
        self._y_label = y_label
        self._filter = filter
        self._plot_data = self._get_plot_data(generations)

    def _get_plot_data(self, generations):
        df = pd.DataFrame()
        self._num_generations = 0
        for id_, generation in enumerate(generations):
            self._num_generations += 1

            filtered = filter(
                self._filter,
                generation.get_molecule_records(),
            )
            properties = tuple(map(self._get_property, filtered))

            # If there are no values after filtering, don't plot
            # anything for the generation.
            if not properties:
                continue

            data = [
                {
                    'Generation': id_,
                    self._y_label: max(properties),
                    'Type': 'Max'
                },
                {
                    'Generation': id_,
                    self._y_label: np.mean(properties),
                    'Type': 'Mean'
                },
                {
                    'Generation': id_,
                    self._y_label: min(properties),
                    'Type': 'Min'
                },
            ]

            df = df.append(data, ignore_index=True)
        return df

    def get_plot_data(self):
        """
        Get the plot data.

        Returns
        -------
        :class:`pandas.DataFrame`
            A data frame holding the plot data.

        """

        return self._plot_data.copy()

    def write(self, path, dpi=500):
        """
        Write a progress plot to a file.

        Parameters
        ----------
        path : :class:`str`
            The path into which the plot is written.

        dpi : :class:`int`, optional
            The dpi of the image.

        Returns
        -------
        :class:`.ProgressPlotter`
            The plotter is returned.

        """

        sns.set(style='darkgrid')
        fig = plt.figure(figsize=[8, 4.5])
        palette = sns.color_palette('deep')

        # It's possible that all values were filtered out, and trying
        # to plot an empty dataframe would raise an exception.
        if len(self._plot_data) != 0:
            sns.scatterplot(
                x='Generation',
                y=self._y_label,
                hue='Type',
                palette={
                    'Max': palette[3],
                    'Min': palette[0],
                    'Mean': palette[2]
                },
                data=self._plot_data,
            )
        # Set the length of the axes to account for all generations,
        # as its possible the first or last ones were not included
        # due to being filtered out.
        plt.xlim(0, self._num_generations)

        plt.legend(bbox_to_anchor=(1.15, 1), prop={'size': 9})
        plt.tight_layout()
        fig.savefig(path, dpi=dpi)
        plt.close('all')
        return self



"""
Fitness Calculator
==================

.. toctree::
    :maxdepth: 2

    Fitness Function <stk.ea.fitness_calculators.fitness_function>
    Property Vector <stk.ea.fitness_calculators.property_vector>

"""


class FitnessCalculator:
    """
    Abstract base class for fitness value calculators.

    Examples
    --------
    *Subclass Implementation*

    You only need to implement :meth:`.get_fitness_value`. The source
    cod of any of the classes listed in :mod:`.fitness_calculator`, can
    serve as good examples.

    """

    def get_fitness_value(self, molecule):
        """
        Return the fitness value of `molecule`.

        Parameters
        ----------
        molecule : :class:`.ConstructedMolecule`
            The molecule whose fitness value should be calculated.

        Returns
        -------
        :class:`object`
            The fitness value of `molecule`.

        """

        raise NotImplementedError()



"""
Sum
===

"""

from .fitness_normalizer import FitnessNormalizer


class Sum(FitnessNormalizer):
    """
    Combines fitness value components into a single fitness value.

    Examples
    --------
    *Combining Fitness Value Components Into a Single Value*

    Sometimes, your :class:`.FitnessCalculator` will return a fitness
    value which is a :class:`tuple` of multiple numbers. Each number
    represents a different property of the molecule, which
    contributes to the final fitness value. This fitness normalizer
    combines these fitness value components into a single number, by
    taking their sum.

    .. code-block:: python

        # Create the normalizer.
        sum_normalizer = stk.Sum()

        # Normalize the fitness values. Assume the fitness values in
        # population are
        # (1, 2, 3), (4, 5, 6), (7, 8, 9).
        normalized = tuple(sum_normalizer.normalize(pop))

        # Fitness values in normalized are 6, 15 and 24.

    *Selectively Normalizing Fitness Values*

    Sometimes, you only want to normalize some members of a population,
    for example if some do not have an assigned fitness value,
    because the fitness calculation failed for whatever reason.
    You can use the `filter` parameter to exclude records from the
    normalization

    .. code-block:: python

        import stk

        normalizer = stk.Sum(
            # Only normalize values which are not None.
            filter=lambda population, record:
                record.get_fitness_value() is not None,
        )
        normalized = tuple(normalizer.normalize(population))

    """

    def __init__(self, filter=lambda population, record: True):
        """
        Initialize a :class:`.Sum` instance.

        Parameters
        ----------
        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._filter = filter

    def normalize(self, population):
        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=sum(record.get_fitness_value()),
                )
            else:
                yield record



"""
M3L6
====

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M3L6(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1, 2)
        | ligands: (3, 4, 5, 6, 7, 8)

    See :class:`.Cage` for more details and examples.

    """

    _R, _theta = 1, 0

    _vertex_prototypes = (
        _NonLinearCageVertex(
            id=0,
            position=[_R*np.cos(_theta), _R*np.sin(_theta), 0]
        ),
        _NonLinearCageVertex(
            id=1,
            position=[
                _R*np.cos(_theta+(4*np.pi/3)),
                _R*np.sin(_theta+(4*np.pi/3)),
                0
            ]
        ),
        _NonLinearCageVertex(
            id=2,
            position=[
                _R*np.cos(_theta+(2*np.pi/3)),
                _R*np.sin(_theta+(2*np.pi/3)),
                0
            ]
        ),

        _LinearCageVertex(
            id=3,
            position=[
                _R*np.cos((_theta+np.pi/4)),
                _R*np.sin((_theta+np.pi/4)),
                0.5
            ],
            use_neighbor_placement=False
        ),
        _LinearCageVertex(
            id=4,
            position=[
                _R*np.cos((_theta+1*np.pi/3)),
                _R*np.sin((_theta+1*np.pi/3)),
                -0.5
            ],
            use_neighbor_placement=False
        ),

        _LinearCageVertex(
            id=5,
            position=[
                _R*np.cos((_theta+1*np.pi/3)+(4*np.pi/3)),
                _R*np.sin((_theta+1*np.pi/3)+(4*np.pi/3)),
                0.5
            ],
            use_neighbor_placement=False
        ),
        _LinearCageVertex(
            id=6,
            position=[
                _R*np.cos((_theta+1*np.pi/3)+(4*np.pi/3)),
                _R*np.sin((_theta+1*np.pi/3)+(4*np.pi/3)),
                -0.5
            ],
            use_neighbor_placement=False
        ),

        _LinearCageVertex(
            id=7,
            position=[
                _R*np.cos((_theta+1*np.pi/3)+(2*np.pi/3)),
                _R*np.sin((_theta+1*np.pi/3)+(2*np.pi/3)),
                0.5
            ],
            use_neighbor_placement=False
        ),
        _LinearCageVertex(
            id=8,
            position=[
                _R*np.cos((_theta+1*np.pi/3)+(2*np.pi/3)),
                _R*np.sin((_theta+1*np.pi/3)+(2*np.pi/3)),
                -0.5
            ],
            use_neighbor_placement=False
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[6]),

        Edge(4, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[6]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[7]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[8]),

        Edge(8, _vertex_prototypes[2], _vertex_prototypes[3]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[4]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[7]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[8]),
    )

    _num_windows = 2
    _num_window_types = 1



"""
Octahedral
==========

"""

from ..metal_complex import MetalComplex
from ..vertices import _MetalVertex, _MonoDentateLigandVertex
from ...topology_graph import Edge


class Octahedral(MetalComplex):
    """
    Represents a metal complex topology graph.

    Metal building blocks with at least six functional groups are
    required for this topology.

    Ligand building blocks with one functional group are required for
    this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, )
        | ligands: (0, 1, 2, 3, 4, 5)

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (
        _MetalVertex(0, [0, 0, 0]),
    )
    _ligand_vertex_prototypes = (
        _MonoDentateLigandVertex(1, [2.5, 0, 0]),
        _MonoDentateLigandVertex(2, [0, 2.5, 0]),
        _MonoDentateLigandVertex(3, [0, 0, 2.5]),
        _MonoDentateLigandVertex(4, [-2.5, 0, 0]),
        _MonoDentateLigandVertex(5, [0, -2.5, 0]),
        _MonoDentateLigandVertex(6, [0, 0, -2.5]),
    )

    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[1],
        ),
        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[2],
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[3],
        ),
        Edge(
            id=4,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[4],
        ),
        Edge(
            id=5,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[5],
        ),
    )



"""
Ten Plus Twenty
===============

"""

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class TenPlusTwenty(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1.5
    _z = _x / 2
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-_x, _x, -_z]),
        _NonLinearCageVertex(1, [-_x, -_x, -_z]),
        _NonLinearCageVertex(2, [_x, _x, -_z]),
        _NonLinearCageVertex(3, [_x, -_x, -_z]),

        _NonLinearCageVertex(4, [-_x, _x, _z]),
        _NonLinearCageVertex(5, [-_x, -_x, _z]),
        _NonLinearCageVertex(6, [_x, _x, _z]),
        _NonLinearCageVertex(7, [_x, -_x, _z]),

        _NonLinearCageVertex(8, [0, 0, _x]),
        _NonLinearCageVertex(9, [0, 0, -_x]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _LinearCageVertex.init_at_center(
            id=10,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=11,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCageVertex.init_at_center(
            id=12,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
        ),
        _LinearCageVertex.init_at_center(
            id=13,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[3]),
        ),

        _LinearCageVertex.init_at_center(
            id=14,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[6]),
        ),
        _LinearCageVertex.init_at_center(
            id=15,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[5]),
        ),
        _LinearCageVertex.init_at_center(
            id=16,
            vertices=(_vertex_prototypes[5], _vertex_prototypes[7]),
        ),
        _LinearCageVertex.init_at_center(
            id=17,
            vertices=(_vertex_prototypes[6], _vertex_prototypes[7]),
        ),

        _LinearCageVertex.init_at_center(
            id=18,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[4]),
        ),
        _LinearCageVertex.init_at_center(
            id=19,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[5]),
        ),
        _LinearCageVertex.init_at_center(
            id=20,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[6]),
        ),
        _LinearCageVertex.init_at_center(
            id=21,
            vertices=(_vertex_prototypes[3], _vertex_prototypes[7]),
        ),

        _LinearCageVertex.init_at_center(
            id=22,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[4]),
        ),
        _LinearCageVertex.init_at_center(
            id=23,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[5]),
        ),
        _LinearCageVertex.init_at_center(
            id=24,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[6]),
        ),
        _LinearCageVertex.init_at_center(
            id=25,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[7]),
        ),

        _LinearCageVertex.init_at_center(
            id=26,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[0]),
        ),
        _LinearCageVertex.init_at_center(
            id=27,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[1]),
        ),
        _LinearCageVertex.init_at_center(
            id=28,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=29,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[3]),
        )
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[10], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[11], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[11], _vertex_prototypes[1]),

        Edge(4, _vertex_prototypes[12], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[12], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[13], _vertex_prototypes[2]),
        Edge(7, _vertex_prototypes[13], _vertex_prototypes[3]),

        Edge(8, _vertex_prototypes[14], _vertex_prototypes[4]),
        Edge(9, _vertex_prototypes[14], _vertex_prototypes[6]),
        Edge(10, _vertex_prototypes[15], _vertex_prototypes[4]),
        Edge(11, _vertex_prototypes[15], _vertex_prototypes[5]),

        Edge(12, _vertex_prototypes[16], _vertex_prototypes[5]),
        Edge(13, _vertex_prototypes[16], _vertex_prototypes[7]),
        Edge(14, _vertex_prototypes[17], _vertex_prototypes[6]),
        Edge(15, _vertex_prototypes[17], _vertex_prototypes[7]),

        Edge(16, _vertex_prototypes[18], _vertex_prototypes[0]),
        Edge(17, _vertex_prototypes[18], _vertex_prototypes[4]),
        Edge(18, _vertex_prototypes[19], _vertex_prototypes[1]),
        Edge(19, _vertex_prototypes[19], _vertex_prototypes[5]),

        Edge(20, _vertex_prototypes[20], _vertex_prototypes[2]),
        Edge(21, _vertex_prototypes[20], _vertex_prototypes[6]),
        Edge(22, _vertex_prototypes[21], _vertex_prototypes[3]),
        Edge(23, _vertex_prototypes[21], _vertex_prototypes[7]),

        Edge(24, _vertex_prototypes[22], _vertex_prototypes[8]),
        Edge(25, _vertex_prototypes[22], _vertex_prototypes[4]),
        Edge(26, _vertex_prototypes[23], _vertex_prototypes[8]),
        Edge(27, _vertex_prototypes[23], _vertex_prototypes[5]),

        Edge(28, _vertex_prototypes[24], _vertex_prototypes[8]),
        Edge(29, _vertex_prototypes[24], _vertex_prototypes[6]),
        Edge(30, _vertex_prototypes[25], _vertex_prototypes[8]),
        Edge(31, _vertex_prototypes[25], _vertex_prototypes[7]),

        Edge(32, _vertex_prototypes[26], _vertex_prototypes[9]),
        Edge(33, _vertex_prototypes[26], _vertex_prototypes[0]),
        Edge(34, _vertex_prototypes[27], _vertex_prototypes[9]),
        Edge(35, _vertex_prototypes[27], _vertex_prototypes[1]),

        Edge(36, _vertex_prototypes[28], _vertex_prototypes[9]),
        Edge(37, _vertex_prototypes[28], _vertex_prototypes[2]),
        Edge(38, _vertex_prototypes[29], _vertex_prototypes[9]),
        Edge(39, _vertex_prototypes[29], _vertex_prototypes[3]),
    )

    _num_windows = 12
    _num_window_types = 2



"""
Thiol Factory
=============

"""

from .functional_group_factory import FunctionalGroupFactory
from .utilities import _get_atom_ids
from ..functional_groups import Thiol


class ThiolFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Thiol` instances.

    Creates functional groups from substructures, which match the
    ``[*][S][H]`` functional group string.

    Examples
    --------
    You want to create a building block which has :class:`.Thiol`
    functional groups. You want the sulfur atom in those functional
    groups to be the *bonder* atom, and the hydrogen atom to be the
    *deleter* atom.

    .. code-block:: python

        import stk

        building_block = stk.BuildingBlock(
            smiles='SCCCCS',
            functional_groups=(stk.ThiolFactory(), ),
        )

    You want to create a building block which has :class:`.Thiol`
    functional groups. You want the non-hydrogen atom bonded to the
    sulfur to be the *bonder* atom and the SH group to be *deleter*
    atoms.

    .. code-block:: python

        import stk

        thiol_factory = stk.ThiolFactory(
            # The index of the atom bonded to sulfur is 0 in the
            # functional group string (see docstring).
            bonders=(0, ),
            # The indices of the sulfur and hydrogen atoms in the
            # functional group string (see docstring) is 1 and 2,
            # respectively.
            deleters=(1, 2),
        )
        building_block = stk.BuildingBlock(
            smiles='SCCCCS',
            functional_groups=(thiol_factory, ),
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1, ), deleters=(2, ), placers=None):
        """
        Initialize a :class:`.ThiolFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in _get_atom_ids('[*][S][H]', molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Thiol(
                sulfur=atoms[1],
                hydrogen=atoms[2],
                atom=atoms[0],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
Shift Up
========

"""

import numpy as np
from functools import partial

from .fitness_normalizer import FitnessNormalizer


class ShiftUp(FitnessNormalizer):
    """
    Shifts negative fitness values to be positive.

    Assume you have a vector-valued fitness value, where each number
    represents a different property of the molecule, for example,
    ``[1, -10, 1]``

    One way to convert the vector-valued fitness value into a
    scalar fitness value is by summing the elements, and the result in
    this case would be ``-8``. Clearly this doesn't work, because the
    resulting fitness value is not a positive number. To fix this,
    the ``-10`` should be shifted to a positive value.

    :class:`.ShiftUp` finds the minimum value of each element in the
    vector-valued fitness value across the entire population, and for
    elements where this minimum value is less than ``0``, shifts up
    the element value for every molecule in the population, so that the
    minimum value in the entire population is ``1``.

    For example, take a population with the vector-valued fitness
    values

    .. code-block:: python

            [1, -5, 5]
            [3, -10, 2]
            [2, 20, 1]
        }

    After normalization the fitness values will be.

    .. code-block:: python

            [1, 6, 5]
            [3, 1, 2]
            [2, 31, 1]

    :class:`.ShiftUp` also works when the fitness value is a
    single value.

    Examples
    --------
    *Ensuring Positive Fitness Values*

    Here you final fitness value is calculated by taking a
    :class:`.Sum` of the different components of the fitness value.
    To ensure that the final sum is positive, each component must
    also be positive.

    .. code-block:: python

        # Create the normalizer.
        shifter = ShiftUp()

        # Normalize the fitness values. Assume the fitness values in
        # population are (1, -2, 3), (4, 5, -6), (7, 8, 9).
        normalized = tuple(shifter.normalize(population))

        # Fitness values in normalized are
        # (1, 1, 10), (4, 8, 1), (7, 11, 16).

    """

    def __init__(self, filter=lambda population, record: True):
        """
        Initialize a :class:`.ShiftUp` instance.

        Parameters
        ----------
        filter : :class:`callable`, optional
            Takes two parameters, first is a :class:`tuple`
            of :class:`.MoleculeRecord` instances,
            and the second is a :class:`.MoleculeRecord`. The
            :class:`callable` returns ``True`` or ``False``. Only
            molecules which return ``True`` will have fitness values
            normalized. By default, all molecules will have fitness
            values normalized.
            The instance passed to the `population` argument of
            :meth:`.normalize` is passed as the first argument, while
            the second argument will be passed every
            :class:`.MoleculeRecord` in it, one at a time.

        """

        self._filter = filter

    def normalize(self, population):
        filtered = filter(
            partial(self._filter, population),
            population,
        )
        # Get all the fitness arrays in a matrix.
        fmat = np.array([
            record.get_fitness_value() for record in filtered
        ])

        # Get the minimum value of each element across the population.
        # keepdims ensures that np.min returns a 1-D array, because
        # it will be True if fitness values are scalar and False if
        # they are array-valued.
        mins = np.min(fmat, axis=0, keepdims=len(fmat.shape) == 1)

        # Convert all elements in mins which are not to be shifted to 0
        # and make the shift equal to the minimum value + 1.
        shift = np.zeros(len(mins))
        for i, min_ in enumerate(mins):
            if min_ <= 0:
                shift[i] = 1 - min_

        for record in population:
            if self._filter(population, record):
                yield record.with_fitness_value(
                    fitness_value=record.get_fitness_value() + shift,
                )
            else:
                yield record



from ......bonds import BondInfo


class _BondBatch:
    """
    A batch of bonds.

    """

    __slots__ = ['_bonds', '_bond_infos']

    def __init__(
        self,
        bonds,
        atom_map,
        building_block,
        building_block_id,
    ):
        """
        Initialize a :class:`.BondBatch` instance.

        Parameters
        ----------
        bonds : :class:`iterable` of :class:`.Bond`
            The bonds, which should be added to the batch.

        atom_map : :class:`dict`
            Maps the ids of atoms held by `bonds`, to the new atoms,
            which the bonds in the :class:`.BondBatchData` instance
            should hold.

        building_block : :class:`.BuildingBlock`
            The building block from which the bonds originate.

        building_block_id : :class:`.int`
            An id, unique to that building block and placement.

        """

        self._bonds = _bonds = []
        self._bond_infos = bond_infos = []

        for bond in bonds:
            _bonds.append(bond.with_atoms(atom_map))
            bond_infos.append(
                BondInfo(
                    bond=_bonds[-1],
                    building_block=building_block,
                    building_block_id=building_block_id,
                )
            )

    def get_bonds(self):
        """
        Yield the bonds in the batch.

        Yields
        ------
        :class:`.Bond`
            A bond in the batch.

        """

        yield from self._bonds

    def get_bond_infos(self):
        """
        Yield info about the bonds in the batch.

        Yields
        ------
        :class:`.BondInfo`
            Info about a bond in the batch.

        """

        yield from self._bond_infos



"""
Construction Result
===================

"""


class ConstructionResult:
    """
    The result of :meth:`.TopologyGraph.construct`.

    """

    __slots__ = [
        '_atoms',
        '_bonds',
        '_atom_infos',
        '_bond_infos',
        '_position_matrix',
    ]

    def __init__(self, construction_state):
        """
        Initialize a :class:`.ConstructionResult`.

        Parameters
        ----------
        construction_state : :class:`.ConstructionState`
            The state from which the result is initialized.

        """

        self._position_matrix = (
            construction_state.get_position_matrix()
        )
        self._position_matrix.setflags(write=False)
        self._atoms = tuple(construction_state.get_atoms())
        self._bonds = tuple(construction_state.get_bonds())
        self._atom_infos = tuple(construction_state.get_atom_infos())
        self._bond_infos = tuple(construction_state.get_bond_infos())

    def get_position_matrix(self):
        """
        Get the position matrix of the constructed molecule.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix of the constructed molecule.

        """

        return self._position_matrix

    def get_atoms(self):
        """
        Get the atoms of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.Atom`
            The atoms of the constructed molecule.

        """

        return self._atoms

    def get_bonds(self):
        """
        Get the bonds of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.Bond`
            The bonds of the constructed molecule.

        """

        return self._bonds

    def get_atom_infos(self):
        """
        Get the atom infos of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.AtomInfo`
            The atom infos of the constructed molecule.

        """

        return self._atom_infos

    def get_bond_infos(self):
        """
        Get the bond infos of the constructed molecule.

        Returns
        -------
        :class:`tuple` of :class:`.BondInfo`
            The bond infos of the constructed molecule.

        """

        return self._bond_infos



"""
Construction State
==================

"""

from .graph_state import _GraphState
from .molecule_state import _MoleculeState


class ConstructionState:
    """
    The state of the molecule and topology graph under construction.

    """

    def __init__(
        self,
        building_block_vertices,
        edges,
        lattice_constants=(),
    ):
        """
        Initialize a :class:`.ConstructionState` instance.

        Parameters
        ----------
        building_block_vertices : :class:`dict`
            Maps each :class:`.BuildingBlock` to be placed, to a
            :class:`tuple` of :class:`.Vertex` instances, on which
            it should be placed.

        edges : :class:`tuple` of :class:`.Edge`
            The edges of the topology graph.

        lattice_constants : :class:`tuple`, optional
            A :class:`numpy.ndarray` for each lattice constant.
            Can be an empty :class:`tuple` if the topology graph is
            not periodic.

        """

        self._graph_state = _GraphState(
            building_block_vertices=building_block_vertices,
            edges=edges,
            lattice_constants=lattice_constants,
        )
        self._molecule_state = _MoleculeState()

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.ConstructionState`
            The clone. Has the same type as the original instance.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._graph_state = self._graph_state
        clone._molecule_state = self._molecule_state
        return clone

    def _with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Modify the instance.

        """

        self._molecule_state = (
            self._molecule_state.with_placement_results(
                vertices=vertices,
                edges=edges,
                building_blocks=building_blocks,
                results=results,
            )
        )
        return self

    def with_placement_results(
        self,
        vertices,
        edges,
        building_blocks,
        results,
    ):
        """
        Return a clone holding the placement results.

        Parameters
        ----------
        vertices : :class:`tuple` of :class:`.Vertex`
            The vertices used for placement.

        edges : :class:`tuple`
            For each vertex in `vertices`, a :class:`tuple` of
            :class:`.Edge` instances connected to it.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            For each vertex in `vertices`, the building block placed
            on it.

        results : :class:`tuple` of :class:`._PlacementResult`
            For every vertex in `vertices`, the result of the
            placement.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the placement results. Has the same
            type as the original instance.

        """

        return self.clone()._with_placement_results(
            vertices=vertices,
            edges=edges,
            building_blocks=building_blocks,
            results=results,
        )

    def get_building_block(self, vertex_id):
        """
        Get the building block to be placed on a given vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of the vertex, on which the building block is to
            be placed.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        return self._graph_state.get_building_block(vertex_id)

    def get_vertices(self, vertex_ids):
        """
        Get the building block to be placed on a given vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of the vertex, on which the building block is to
            be placed.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        yield from self._graph_state.get_vertices(vertex_ids)

    def get_num_vertices(self):
        """
        Get the number of vertices in the topology graph.

        Returns
        -------
        :class:`int`
            The number of vertices in the topology graph.

        """

        return self._graph_state.get_num_vertices()

    def get_edge(self, edge_id):
        """
        Get an edge.

        Parameters
        ----------
        edge_id : :class:`int`
            The id of an edge.

        Returns
        -------
        :class:`.Edge`
            An edge.

        """

        return self._graph_state.get_edge(edge_id)

    def get_num_edges(self):
        """
        Get the number of edges in the topology graph.

        Returns
        -------
        :class:`int`
            The number of edges.

        """

        return self._graph_state.get_num_edges()

    def get_edges(self, vertex_id):
        """
        Get the edges connect to a vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of a vertex.

        Returns
        -------
        :class:`tuple` of :class:`.Edge`
            The connected edges.

        """

        return self._graph_state.get_edges(vertex_id)

    def get_edge_group_functional_groups(self, edge_group):
        """
        Yield the functional groups associated with `edge_group`.

        Parameters
        ----------
        edge_group : :class:`.EdgeGroup`
            The edge group, whose functional groups are desired.

        Yields
        ------
        :class:`.FunctionalGroup`
            A functional group which belongs to `edge_group`.

        """

        yield from (
            self._molecule_state.get_edge_group_functional_groups(
                edge_group=edge_group,
            )
        )

    def _with_reaction_results(self, reactions, results):
        """
        Modify the instance.

        """

        self._molecule_state = (
            self._molecule_state.with_reaction_results(
                reactions=reactions,
                results=results,
            )
        )
        return self

    def with_reaction_results(self, reactions, results):
        """
        Return a clone holding the reaction results.

        Parameters
        ----------
        reactions : :class:`tuple` of :class:`.Reaction`
            The reactions.

        results : :class:`.ReactionResult`
            For each reaction in `reactions`, its result.

        Returns
        -------
        :class:`.ConstructionState`
            The clone holding the reaction results. Has the same type
            as the original instance.

        """

        return self.clone()._with_reaction_results(reactions, results)

    def _with_vertices(self, vertices):
        """
        Modify the instance.

        """

        self._graph_state = self._graph_state.with_vertices(vertices)
        return self

    def with_vertices(self, vertices):
        """
        Returns a clone holding `vertices`.

        Parameters
        ----------
        vertices : :class:`iterable` of :class:`.Vertex`
            The vertices the clone should hold.

        Returns
        -------
        :class:`.ConstructionState`
            The clone. Has the same type as the original instance.

        """

        return self.clone()._with_vertices(vertices)

    def get_position_matrix(self):
        """
        Get the position matrix of the molecule being constructed.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix.

        """

        return self._molecule_state.get_position_matrix()

    def get_atoms(self):
        """
        Yield the atoms of the molecule being constructed.

        Yields
        ------
        :class:`.Atom`
            An atom of the molecule being constructed.

        """

        yield from self._molecule_state.get_atoms()

    def get_bonds(self):
        """
        Yield the bonds of the molecule being constructed.

        Yields
        ------
        :class:`.Bond`
            A bond of the molecule being constructed.

        """

        yield from self._molecule_state.get_bonds()

    def get_atom_infos(self):
        """
        Yield the atom infos of the molecule being constructed.

        Yields
        ------
        :class:`.AtomInfo`
            An atom info of the molecule being constructed.

        """

        yield from self._molecule_state.get_atom_infos()

    def get_bond_infos(self):
        """
        Yield the bond infos of the molecule being constructed.

        Yields
        ------
        :class:`.BondInfo`
            The bond info of the molecule being constructed.

        """

        yield from self._molecule_state.get_bond_infos()



"""
Four Plus Six
=============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class FourPlusSix(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    # Vertices of a tetrahdron so that origin is at the origin. Source:
    # http://tinyurl.com/lc262h8.
    _v0, _v1, _v2, _v3 = _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0, np.sqrt(6)/2]),
        _NonLinearCageVertex(1, [-1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(2, [1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(3, [0, 2*np.sqrt(3)/3, -np.sqrt(6)/6]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCageVertex.init_at_center(4, (_v0, _v1)),
        _LinearCageVertex.init_at_center(5, (_v0, _v2)),
        _LinearCageVertex.init_at_center(6, (_v0, _v3)),
        _LinearCageVertex.init_at_center(7, (_v1, _v2)),
        _LinearCageVertex.init_at_center(8, (_v1, _v3)),
        _LinearCageVertex.init_at_center(9, (_v2, _v3)),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[6], _vertex_prototypes[0]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[2]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[1]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[3]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[2]),
        Edge(11, _vertex_prototypes[9], _vertex_prototypes[3])
    )

    _num_windows = 4
    _num_window_types = 1



from ......atoms import AtomInfo


class _AtomBatch:
    """
    A batch of atoms.

    """

    __slots__ = ['_atoms', '_atom_infos', '_atom_map']

    def __init__(
        self,
        atoms,
        num_atoms,
        building_block,
        building_block_id,
    ):
        """
        Initialize an :class:`._AtomBatch` instance.

        Parameters
        ----------
        atoms : :class:`iterable` of :class:`.Atom`
            The atoms, which should be added to the batch.

        num_atoms : :class:`int`
            The number of atoms in the molecule being constructed,
            before atoms in this batch are taken into account.

        building_block : :class:`.BuildingBlock`
            The building block from which the atoms originate.

        building_block_id : :class:`.int`
            An id, unique to that building block and placement.

        """

        self._atoms = _atoms = []
        self._atom_infos = atom_infos = []
        self._atom_map = atom_map = {}

        for id_, atom in enumerate(atoms, num_atoms):
            _atoms.append(atom.with_id(id_))
            atom_map[atom.get_id()] = _atoms[-1]
            atom_infos.append(
                AtomInfo(
                    atom=_atoms[-1],
                    building_block=building_block,
                    building_block_id=building_block_id,
                )
            )

    def get_atoms(self):
        """
        Yield the atoms in the batch.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        yield from self._atoms

    def get_atom_infos(self):
        """
        Yield info about the atoms in the batch.

        Yields
        ------
        :class:`.AtomInfo`
            Info about an atom in the batch.

        """

        yield from self._atom_infos

    def get_atom_map(self):
        """
        Get a mapping from the old atom id to the new atom.

        Returns
        -------
        :class:`dict`
            Maps the id of an atom provided to the initializer, to
            the new atom held by the batch, which has an updated id.

        """

        return dict(self._atom_map)



"""
M6L2L3 Prism
============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class M6L2L3Prism(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with three functional groups are
    required for this topology.

    Ligand building blocks with three and four functional groups are
    required for ligand type A and B, respectively, on this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: 0 to 5
        | ligand A (3 functional groups): (6, 7)
        | ligand B (4 functional groups): (8, 9, 10)

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-1, -1/np.sqrt(3), 1]),
        _NonLinearCageVertex(1, [1, -1/np.sqrt(3), 1]),
        _NonLinearCageVertex(2, [0, 2/np.sqrt(3), 1]),

        _NonLinearCageVertex(3, [-1, -1/np.sqrt(3), -1]),
        _NonLinearCageVertex(4, [1, -1/np.sqrt(3), -1]),
        _NonLinearCageVertex(5, [0, 2/np.sqrt(3), -1]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _NonLinearCageVertex.init_at_center(
            id=6,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[1],
                _vertex_prototypes[2],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=7,
            vertices=(
                _vertex_prototypes[3],
                _vertex_prototypes[4],
                _vertex_prototypes[5],
            ),
        ),

        _NonLinearCageVertex.init_at_center(
            id=8,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[1],
                _vertex_prototypes[3],
                _vertex_prototypes[4],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=9,
            vertices=(
                _vertex_prototypes[1],
                _vertex_prototypes[2],
                _vertex_prototypes[4],
                _vertex_prototypes[5],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=10,
            vertices=(
                _vertex_prototypes[2],
                _vertex_prototypes[0],
                _vertex_prototypes[5],
                _vertex_prototypes[3],
            ),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[10]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[6]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[9]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(7, _vertex_prototypes[2], _vertex_prototypes[9]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[3], _vertex_prototypes[10]),
        Edge(12, _vertex_prototypes[4], _vertex_prototypes[7]),
        Edge(13, _vertex_prototypes[4], _vertex_prototypes[8]),
        Edge(14, _vertex_prototypes[4], _vertex_prototypes[9]),
        Edge(15, _vertex_prototypes[5], _vertex_prototypes[7]),
        Edge(16, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(17, _vertex_prototypes[5], _vertex_prototypes[10]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
Elements
========

Defines an :class:`.Atom` class for each element.

"""

from .atom import Atom


class H(Atom):
    _atomic_number = 1


class He(Atom):
    _atomic_number = 2


class Li(Atom):
    _atomic_number = 3


class Be(Atom):
    _atomic_number = 4


class B(Atom):
    _atomic_number = 5


class C(Atom):
    _atomic_number = 6


class N(Atom):
    _atomic_number = 7


# "O" is a valid elemental symbol.
class O(Atom):  # noqa
    _atomic_number = 8


class F(Atom):
    _atomic_number = 9


class Ne(Atom):
    _atomic_number = 10


class Na(Atom):
    _atomic_number = 11


class Mg(Atom):
    _atomic_number = 12


class Al(Atom):
    _atomic_number = 13


class Si(Atom):
    _atomic_number = 14


class P(Atom):
    _atomic_number = 15


class S(Atom):
    _atomic_number = 16


class Cl(Atom):
    _atomic_number = 17


class Ar(Atom):
    _atomic_number = 18


class K(Atom):
    _atomic_number = 19


class Ca(Atom):
    _atomic_number = 20


class Sc(Atom):
    _atomic_number = 21


class Ti(Atom):
    _atomic_number = 22


class V(Atom):
    _atomic_number = 23


class Cr(Atom):
    _atomic_number = 24


class Mn(Atom):
    _atomic_number = 25


class Fe(Atom):
    _atomic_number = 26


class Co(Atom):
    _atomic_number = 27


class Ni(Atom):
    _atomic_number = 28


class Cu(Atom):
    _atomic_number = 29


class Zn(Atom):
    _atomic_number = 30


class Ga(Atom):
    _atomic_number = 31


class Ge(Atom):
    _atomic_number = 32


class As(Atom):
    _atomic_number = 33


class Se(Atom):
    _atomic_number = 34


class Br(Atom):
    _atomic_number = 35


class Kr(Atom):
    _atomic_number = 36


class Rb(Atom):
    _atomic_number = 37


class Sr(Atom):
    _atomic_number = 38


class Y(Atom):
    _atomic_number = 39


class Zr(Atom):
    _atomic_number = 40


class Nb(Atom):
    _atomic_number = 41


class Mo(Atom):
    _atomic_number = 42


class Tc(Atom):
    _atomic_number = 43


class Ru(Atom):
    _atomic_number = 44


class Rh(Atom):
    _atomic_number = 45


class Pd(Atom):
    _atomic_number = 46


class Ag(Atom):
    _atomic_number = 47


class Cd(Atom):
    _atomic_number = 48


class In(Atom):
    _atomic_number = 49


class Sn(Atom):
    _atomic_number = 50


class Sb(Atom):
    _atomic_number = 51


class Te(Atom):
    _atomic_number = 52


# "I" is a valid elemental symbol.
class I(Atom):  # noqa
    _atomic_number = 53


class Xe(Atom):
    _atomic_number = 54


class Cs(Atom):
    _atomic_number = 55


class Ba(Atom):
    _atomic_number = 56


class La(Atom):
    _atomic_number = 57


class Ce(Atom):
    _atomic_number = 58


class Pr(Atom):
    _atomic_number = 59


class Nd(Atom):
    _atomic_number = 60


class Pm(Atom):
    _atomic_number = 61


class Sm(Atom):
    _atomic_number = 62


class Eu(Atom):
    _atomic_number = 63


class Gd(Atom):
    _atomic_number = 64


class Tb(Atom):
    _atomic_number = 65


class Dy(Atom):
    _atomic_number = 66


class Ho(Atom):
    _atomic_number = 67


class Er(Atom):
    _atomic_number = 68


class Tm(Atom):
    _atomic_number = 69


class Yb(Atom):
    _atomic_number = 70


class Lu(Atom):
    _atomic_number = 71


class Hf(Atom):
    _atomic_number = 72


class Ta(Atom):
    _atomic_number = 73


class W(Atom):
    _atomic_number = 74


class Re(Atom):
    _atomic_number = 75


class Os(Atom):
    _atomic_number = 76


class Ir(Atom):
    _atomic_number = 77


class Pt(Atom):
    _atomic_number = 78


class Au(Atom):
    _atomic_number = 79


class Hg(Atom):
    _atomic_number = 80


class Tl(Atom):
    _atomic_number = 81


class Pb(Atom):
    _atomic_number = 82


class Bi(Atom):
    _atomic_number = 83


class Po(Atom):
    _atomic_number = 84


class At(Atom):
    _atomic_number = 85


class Rn(Atom):
    _atomic_number = 86


class Fr(Atom):
    _atomic_number = 87


class Ra(Atom):
    _atomic_number = 88


class Ac(Atom):
    _atomic_number = 89


class Th(Atom):
    _atomic_number = 90


class Pa(Atom):
    _atomic_number = 91


class U(Atom):
    _atomic_number = 92


class Np(Atom):
    _atomic_number = 93


class Pu(Atom):
    _atomic_number = 94


class Am(Atom):
    _atomic_number = 95


class Cm(Atom):
    _atomic_number = 96


class Bk(Atom):
    _atomic_number = 97


class Cf(Atom):
    _atomic_number = 98


class Es(Atom):
    _atomic_number = 99


class Fm(Atom):
    _atomic_number = 100


class Md(Atom):
    _atomic_number = 101


class No(Atom):
    _atomic_number = 102


class Lr(Atom):
    _atomic_number = 103


class Rf(Atom):
    _atomic_number = 104


class Db(Atom):
    _atomic_number = 105


class Sg(Atom):
    _atomic_number = 106


class Bh(Atom):
    _atomic_number = 107


class Hs(Atom):
    _atomic_number = 108


class Mt(Atom):
    _atomic_number = 109


class Ds(Atom):
    _atomic_number = 110


class Rg(Atom):
    _atomic_number = 111


class Cn(Atom):
    _atomic_number = 112


class Nh(Atom):
    _atomic_number = 113


class Fl(Atom):
    _atomic_number = 114


class Mc(Atom):
    _atomic_number = 115


class Lv(Atom):
    _atomic_number = 116


class Ts(Atom):
    _atomic_number = 117


class Og(Atom):
    _atomic_number = 118



"""
Linkerless Honeycomb
====================

"""

import numpy as np

from .cof import Cof
from .vertices import _NonLinearCofVertex
from ..topology_graph import Edge


class LinkerlessHoneycomb(Cof):
    """
    Represents a honeycomb COF topology graph.

    Building blocks with three functional groups are required
    for this topology graph.

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0.]),
        np.array([0., 0., 5/1.7321]),
    )

    _vertex_prototypes = (
        _NonLinearCofVertex(0, (1/3)*_a + (1/3)*_b + (1/2)*_c),
        _NonLinearCofVertex(1, (2/3)*_a + (2/3)*_b + (1/2)*_c),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(-1, 0, 0),
        ),
        Edge(
            id=2,
            vertex1=_vertex_prototypes[0],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),
    )



"""
Normalizer Sequence
===================

"""

from .fitness_normalizer import FitnessNormalizer


class NormalizerSequence(FitnessNormalizer):
    def __init__(self, fitness_normalizers):
        self._fitness_normalizers = fitness_normalizers

    def normalize(self, population):
        for normalizer in self._fitness_normalizers:
            population = tuple(normalizer.normalize(population))
        yield from population



"""
Four Plus Eight
===============

"""

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class FourPlusEight(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-1, -1, 0], False),
        _NonLinearCageVertex(1, [-1, 1, 0], False),

        _NonLinearCageVertex(2, [1, -1, 0], False),
        _NonLinearCageVertex(3, [1, 1, 0], False),

        _LinearCageVertex(4, [-2, 0, 1], False),
        _LinearCageVertex(5, [-2, 0, -1], False),

        _LinearCageVertex(6, [0, 2, 1], False),
        _LinearCageVertex(7, [0, 2, -1], False),

        _LinearCageVertex(8, [0, -2, 1], False),
        _LinearCageVertex(9, [0, -2, -1], False),

        _LinearCageVertex(10, [2, 0, 1], False),
        _LinearCageVertex(11, [2, 0, -1], False),

    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),

        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[1]),

        Edge(4, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[3]),

        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),

        Edge(8, _vertex_prototypes[8], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[2]),

        Edge(10, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(11, _vertex_prototypes[9], _vertex_prototypes[2]),

        Edge(12, _vertex_prototypes[10], _vertex_prototypes[2]),
        Edge(13, _vertex_prototypes[10], _vertex_prototypes[3]),

        Edge(14, _vertex_prototypes[11], _vertex_prototypes[2]),
        Edge(15, _vertex_prototypes[11], _vertex_prototypes[3]),

    )

    _num_windows = 6
    _num_window_types = 2



"""
Square
======

"""

import numpy as np


from .cof import Cof
from .vertices import _LinearCofVertex, _NonLinearCofVertex
from ..topology_graph import Edge


class Square(Cof):
    """
    Represents a sqaure COF topology graph.

    Building blocks with four and two functional groups are required
    for this topology graph.

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0., 1., 0.]),
        np.array([0., 0., 1.])
    )

    _vertex_prototypes = (
        _NonLinearCofVertex(0, (0.5)*_a + (0.5)*_b + (0.5)*_c),
    )
    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCofVertex.init_at_shifted_center(
            id=1,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[0]),
            cell_shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=2,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[0]),
            cell_shifts=((0, 0, 0), (0, 1, 0)),
            lattice_constants=_lattice_constants,
        ),

    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[1], _vertex_prototypes[0]),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[1],
            vertex2=_vertex_prototypes[0],
            periodicity=(1, 0, 0),
        ),
        Edge(2, _vertex_prototypes[2], _vertex_prototypes[0]),
        Edge(
            id=3,
            vertex1=_vertex_prototypes[2],
            vertex2=_vertex_prototypes[0],
            periodicity=(0, 1, 0),
        ),
    )



"""
Molecule DeJSONizer
===================

"""

import numpy as np
from stk.molecular import Molecule

from .utilities import to_atom, to_bond


class MoleculeDejsonizer:
    """
    Abstract base class for creating molecules from JSONs.

    See Also
    --------
    :class:`.ConstructedMoleculeDejsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    """

    def from_json(self, json):
        """
        Get a :class:`.Molecule` from a JSON.

        Parameters
        ----------
        json : :class:`dict`
            A JSON representation of a molecule.

        Returns
        -------
        :class:`.Molecule`
            The molecule held in `json`.

        """

        atoms = tuple(
            to_atom(atom_id, atom_json)
            for atom_id, atom_json in enumerate(json['molecule']['a'])
        )
        return Molecule(
            atoms=atoms,
            bonds=tuple(
                to_bond(atoms, bond_json)
                for bond_json in json['molecule']['b']
            ),
            position_matrix=np.array(json['matrix']['m']),
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{self.__class__.__name__}()'



"""
Twenty Plus Thirty
==================

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class TwentyPlusThirty(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    # Source: http://tinyurl.com/h2dl949
    _phi = (1 + np.sqrt(5))/2
    _x = 1.5
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [_x*_phi, 0.0, _x/_phi]),
        _NonLinearCageVertex(1, [_x*-_phi, 0.0, _x/_phi]),
        _NonLinearCageVertex(2, [_x*-_phi, 0.0, _x/-_phi]),
        _NonLinearCageVertex(3, [_x*_phi, 0.0, _x/-_phi]),

        _NonLinearCageVertex(4, [_x/_phi, _x*_phi, 0.0]),
        _NonLinearCageVertex(5, [_x/_phi, _x*-_phi, 0.0]),
        _NonLinearCageVertex(6, [_x/-_phi, _x*-_phi, 0.0]),
        _NonLinearCageVertex(7, [_x/-_phi, _x*_phi, 0.0]),
        _NonLinearCageVertex(8, [0.0, _x/_phi, _x*_phi]),
        _NonLinearCageVertex(9, [0.0, _x/_phi, _x*-_phi]),
        _NonLinearCageVertex(10, [0.0, _x/-_phi, _x*-_phi]),
        _NonLinearCageVertex(11, [0.0, _x/-_phi, _x*_phi]),

        _NonLinearCageVertex(12, [_x, _x, _x]),
        _NonLinearCageVertex(13, [_x, -_x, _x]),
        _NonLinearCageVertex(14, [-_x, -_x, _x]),
        _NonLinearCageVertex(15, [-_x, _x, _x]),
        _NonLinearCageVertex(16, [-_x, _x, -_x]),
        _NonLinearCageVertex(17, [_x, _x, -_x]),
        _NonLinearCageVertex(18, [_x, -_x, -_x]),
        _NonLinearCageVertex(19, [-_x, -_x, -_x]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCageVertex.init_at_center(
            id=20,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[13]),
        ),
        _LinearCageVertex.init_at_center(
            id=21,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[12]),
        ),
        _LinearCageVertex.init_at_center(
            id=22,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[3]),
        ),

        _LinearCageVertex.init_at_center(
            id=23,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[14]),
        ),
        _LinearCageVertex.init_at_center(
            id=24,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[15]),
        ),
        _LinearCageVertex.init_at_center(
            id=25,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
        ),

        _LinearCageVertex.init_at_center(
            id=26,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[19]),
        ),
        _LinearCageVertex.init_at_center(
            id=27,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[16]),
        ),

        _LinearCageVertex.init_at_center(
            id=28,
            vertices=(_vertex_prototypes[3], _vertex_prototypes[18]),
        ),
        _LinearCageVertex.init_at_center(
            id=29,
            vertices=(_vertex_prototypes[3], _vertex_prototypes[17]),
        ),

        _LinearCageVertex.init_at_center(
            id=30,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[12]),
        ),
        _LinearCageVertex.init_at_center(
            id=31,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[7]),
        ),
        _LinearCageVertex.init_at_center(
            id=32,
            vertices=(_vertex_prototypes[4], _vertex_prototypes[17]),
        ),

        _LinearCageVertex.init_at_center(
            id=33,
            vertices=(_vertex_prototypes[5], _vertex_prototypes[6]),
        ),
        _LinearCageVertex.init_at_center(
            id=34,
            vertices=(_vertex_prototypes[5], _vertex_prototypes[18]),
        ),
        _LinearCageVertex.init_at_center(
            id=35,
            vertices=(_vertex_prototypes[5], _vertex_prototypes[13]),
        ),

        _LinearCageVertex.init_at_center(
            id=36,
            vertices=(_vertex_prototypes[6], _vertex_prototypes[14]),
        ),
        _LinearCageVertex.init_at_center(
            id=37,
            vertices=(_vertex_prototypes[6], _vertex_prototypes[19]),
        ),

        _LinearCageVertex.init_at_center(
            id=38,
            vertices=(_vertex_prototypes[7], _vertex_prototypes[15]),
        ),
        _LinearCageVertex.init_at_center(
            id=39,
            vertices=(_vertex_prototypes[7], _vertex_prototypes[16]),
        ),

        _LinearCageVertex.init_at_center(
            id=40,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[11]),
        ),
        _LinearCageVertex.init_at_center(
            id=41,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[12]),
        ),
        _LinearCageVertex.init_at_center(
            id=42,
            vertices=(_vertex_prototypes[8], _vertex_prototypes[15]),
        ),

        _LinearCageVertex.init_at_center(
            id=43,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[10]),
        ),
        _LinearCageVertex.init_at_center(
            id=44,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[17]),
        ),
        _LinearCageVertex.init_at_center(
            id=45,
            vertices=(_vertex_prototypes[9], _vertex_prototypes[16]),
        ),

        _LinearCageVertex.init_at_center(
            id=46,
            vertices=(_vertex_prototypes[10], _vertex_prototypes[18]),
        ),
        _LinearCageVertex.init_at_center(
            id=47,
            vertices=(_vertex_prototypes[10], _vertex_prototypes[19]),
        ),

        _LinearCageVertex.init_at_center(
            id=48,
            vertices=(_vertex_prototypes[11], _vertex_prototypes[14]),
        ),
        _LinearCageVertex.init_at_center(
            id=49,
            vertices=(_vertex_prototypes[11], _vertex_prototypes[13])),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[20], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[20], _vertex_prototypes[13]),
        Edge(2, _vertex_prototypes[21], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[21], _vertex_prototypes[12]),
        Edge(4, _vertex_prototypes[22], _vertex_prototypes[0]),
        Edge(5, _vertex_prototypes[22], _vertex_prototypes[3]),

        Edge(6, _vertex_prototypes[23], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[23], _vertex_prototypes[14]),
        Edge(8, _vertex_prototypes[24], _vertex_prototypes[1]),
        Edge(9, _vertex_prototypes[24], _vertex_prototypes[15]),
        Edge(10, _vertex_prototypes[25], _vertex_prototypes[1]),
        Edge(11, _vertex_prototypes[25], _vertex_prototypes[2]),

        Edge(12, _vertex_prototypes[26], _vertex_prototypes[2]),
        Edge(13, _vertex_prototypes[26], _vertex_prototypes[19]),
        Edge(14, _vertex_prototypes[27], _vertex_prototypes[2]),
        Edge(15, _vertex_prototypes[27], _vertex_prototypes[16]),

        Edge(16, _vertex_prototypes[28], _vertex_prototypes[3]),
        Edge(17, _vertex_prototypes[28], _vertex_prototypes[18]),
        Edge(18, _vertex_prototypes[29], _vertex_prototypes[3]),
        Edge(19, _vertex_prototypes[29], _vertex_prototypes[17]),

        Edge(20, _vertex_prototypes[30], _vertex_prototypes[4]),
        Edge(21, _vertex_prototypes[30], _vertex_prototypes[12]),
        Edge(22, _vertex_prototypes[31], _vertex_prototypes[4]),
        Edge(23, _vertex_prototypes[31], _vertex_prototypes[7]),
        Edge(24, _vertex_prototypes[32], _vertex_prototypes[4]),
        Edge(25, _vertex_prototypes[32], _vertex_prototypes[17]),

        Edge(26, _vertex_prototypes[33], _vertex_prototypes[5]),
        Edge(27, _vertex_prototypes[33], _vertex_prototypes[6]),
        Edge(28, _vertex_prototypes[34], _vertex_prototypes[5]),
        Edge(29, _vertex_prototypes[34], _vertex_prototypes[18]),
        Edge(30, _vertex_prototypes[35], _vertex_prototypes[5]),
        Edge(31, _vertex_prototypes[35], _vertex_prototypes[13]),

        Edge(32, _vertex_prototypes[36], _vertex_prototypes[6]),
        Edge(33, _vertex_prototypes[36], _vertex_prototypes[14]),
        Edge(34, _vertex_prototypes[37], _vertex_prototypes[6]),
        Edge(35, _vertex_prototypes[37], _vertex_prototypes[19]),

        Edge(36, _vertex_prototypes[38], _vertex_prototypes[7]),
        Edge(37, _vertex_prototypes[38], _vertex_prototypes[15]),
        Edge(38, _vertex_prototypes[39], _vertex_prototypes[7]),
        Edge(39, _vertex_prototypes[39], _vertex_prototypes[16]),

        Edge(40, _vertex_prototypes[40], _vertex_prototypes[8]),
        Edge(41, _vertex_prototypes[40], _vertex_prototypes[11]),
        Edge(42, _vertex_prototypes[41], _vertex_prototypes[8]),
        Edge(43, _vertex_prototypes[41], _vertex_prototypes[12]),
        Edge(44, _vertex_prototypes[42], _vertex_prototypes[8]),
        Edge(45, _vertex_prototypes[42], _vertex_prototypes[15]),

        Edge(46, _vertex_prototypes[43], _vertex_prototypes[9]),
        Edge(47, _vertex_prototypes[43], _vertex_prototypes[10]),
        Edge(48, _vertex_prototypes[44], _vertex_prototypes[9]),
        Edge(49, _vertex_prototypes[44], _vertex_prototypes[17]),
        Edge(50, _vertex_prototypes[45], _vertex_prototypes[9]),
        Edge(51, _vertex_prototypes[45], _vertex_prototypes[16]),

        Edge(52, _vertex_prototypes[46], _vertex_prototypes[10]),
        Edge(53, _vertex_prototypes[46], _vertex_prototypes[18]),
        Edge(54, _vertex_prototypes[47], _vertex_prototypes[10]),
        Edge(55, _vertex_prototypes[47], _vertex_prototypes[19]),

        Edge(56, _vertex_prototypes[48], _vertex_prototypes[11]),
        Edge(57, _vertex_prototypes[48], _vertex_prototypes[14]),
        Edge(58, _vertex_prototypes[49], _vertex_prototypes[11]),
        Edge(59, _vertex_prototypes[49], _vertex_prototypes[13]),
    )

    _num_windows = 12
    _num_window_types = 1



"""
Five Plus Ten
=============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class FivePlusTen(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _c1 = np.cos(2*np.pi/5)
    _c2 = np.cos(np.pi/5)
    _s1 = np.sin(2*np.pi/5)
    _s2 = np.sin(4*np.pi/5)

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 1, 0], False),
        _NonLinearCageVertex(1, [_s1, _c1, 0], False),
        _NonLinearCageVertex(2, [_s2, -_c2, 0], False),

        _NonLinearCageVertex(3, [-_s2, -_c2, 0], False),
        _NonLinearCageVertex(4, [-_s1, _c1, 0], False),

        _LinearCageVertex(5, [_s1, 1+_c1, 0.5], False),
        _LinearCageVertex(6, [_s1, 1+_c1, -0.5], False),

        _LinearCageVertex(7, [_s1+_s2, _c1-_c2, 0.5], False),
        _LinearCageVertex(8, [_s1+_s2, _c1-_c2, -0.5], False),

        _LinearCageVertex(9, [0, -2*_c2, 0.5], False),
        _LinearCageVertex(10, [0, -2*_c2, -0.5], False),

        _LinearCageVertex(11, [-_s2-_s1, -_c2+_c1, 0.5], False),
        _LinearCageVertex(12, [-_s2-_s1, -_c2+_c1, -0.5], False),

        _LinearCageVertex(13, [-_s1, 1+_c1, 0.5], False),
        _LinearCageVertex(14, [-_s1, 1+_c1, -0.5], False),

    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[5], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[6], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[6], _vertex_prototypes[1]),

        Edge(4, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[7], _vertex_prototypes[2]),
        Edge(6, _vertex_prototypes[8], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[8], _vertex_prototypes[2]),

        Edge(8, _vertex_prototypes[9], _vertex_prototypes[2]),
        Edge(9, _vertex_prototypes[9], _vertex_prototypes[3]),
        Edge(10, _vertex_prototypes[10], _vertex_prototypes[2]),
        Edge(11, _vertex_prototypes[10], _vertex_prototypes[3]),

        Edge(12, _vertex_prototypes[11], _vertex_prototypes[3]),
        Edge(13, _vertex_prototypes[11], _vertex_prototypes[4]),
        Edge(14, _vertex_prototypes[12], _vertex_prototypes[3]),
        Edge(15, _vertex_prototypes[12], _vertex_prototypes[4]),

        Edge(16, _vertex_prototypes[13], _vertex_prototypes[4]),
        Edge(17, _vertex_prototypes[13], _vertex_prototypes[0]),
        Edge(18, _vertex_prototypes[14], _vertex_prototypes[4]),
        Edge(19, _vertex_prototypes[14], _vertex_prototypes[0]),

    )

    _num_windows = 7
    _num_window_types = 2



"""
Hexagonal
=========

"""

import numpy as np


from .cof import Cof
from .vertices import _LinearCofVertex, _NonLinearCofVertex
from ..topology_graph import Edge


class Hexagonal(Cof):
    """
    Represents a hexagonal COF topology graph.

    Building blocks with six and two functional groups are required
    for this topology graph.

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1., 0., 0.]),
        np.array([0.5, 0.866, 0]),
        np.array([0, 0, 5/1.7321])
    )

    _vertex_prototypes = (
        _NonLinearCofVertex(0, (1/4)*_a + (1/4)*_b + (1/2)*_c),
        _NonLinearCofVertex(1, (1/4)*_a + (3/4)*_b + (1/2)*_c),
        _NonLinearCofVertex(2, (3/4)*_a + (1/4)*_b + (1/2)*_c),
        _NonLinearCofVertex(3, (3/4)*_a + (3/4)*_b + (1/2)*_c),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCofVertex.init_at_center(
            id=4,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCofVertex.init_at_center(
            id=5,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCofVertex.init_at_center(
            id=6,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
        ),
        _LinearCofVertex.init_at_center(
            id=7,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
        ),
        _LinearCofVertex.init_at_center(
            id=8,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[3]),
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=9,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=10,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=11,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=12,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[1]),
            cell_shifts=((0, 0, 0), (1, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=13,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[3]),
            cell_shifts=((0, 0, 0), (0, -1, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=14,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
            cell_shifts=((0, 0, 0), (-1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        _LinearCofVertex.init_at_shifted_center(
            id=15,
            vertices=(_vertex_prototypes[3], _vertex_prototypes[0]),
            cell_shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants,
        )
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),

        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[2]),

        Edge(4, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[2]),

        Edge(6, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),

        Edge(8, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[3]),

        Edge(10, _vertex_prototypes[9], _vertex_prototypes[0]),
        Edge(
            id=11,
            vertex1=_vertex_prototypes[9],
            vertex2=_vertex_prototypes[2],
            periodicity=(-1, 0, 0),
        ),

        Edge(12, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(
            id=13,
            vertex1=_vertex_prototypes[10],
            vertex2=_vertex_prototypes[1],
            periodicity=(0, -1, 0),
        ),

        Edge(14, _vertex_prototypes[11], _vertex_prototypes[0]),
        Edge(
            id=15,
            vertex1=_vertex_prototypes[11],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),

        Edge(16, _vertex_prototypes[12], _vertex_prototypes[2]),
        Edge(
            id=17,
            vertex1=_vertex_prototypes[12],
            vertex2=_vertex_prototypes[1],
            periodicity=(1, -1, 0),
        ),

        Edge(18, _vertex_prototypes[13], _vertex_prototypes[2]),
        Edge(
            id=19,
            vertex1=_vertex_prototypes[13],
            vertex2=_vertex_prototypes[3],
            periodicity=(0, -1, 0),
        ),

        Edge(20, _vertex_prototypes[14], _vertex_prototypes[1]),
        Edge(
            id=21,
            vertex1=_vertex_prototypes[14],
            vertex2=_vertex_prototypes[3],
            periodicity=(-1, 0, 0),
        ),

        Edge(22, _vertex_prototypes[15], _vertex_prototypes[3]),
        Edge(
            id=23,
            vertex1=_vertex_prototypes[15],
            vertex2=_vertex_prototypes[0],
            periodicity=(1, 0, 0),
        ),
    )



from stk.utilities import periodic_table


def _with_structure_from_turbomole(self, path):
    """
    Return a clone, with its structure taken from a Turbomole file.

    Note that coordinates in ``.coord`` files are given in Bohr.

    Parameters
    ----------
    path : :class:`str`
        The full path of the ``.coord`` file from which the
        structure should be updated.

    Returns
    -------
    :class:`.Molecule`
        A clone with atomic positions found in `path`.

    Raises
    ------
    :class:`RuntimeError`
        If the number of atoms in the file does not match the
        number of atoms in the molecule or if atom elements in the
        file do not agree with the atom elements in the molecule.

    """

    bohr_to_ang = 0.5291772105638411

    with open(path, 'r') as f:
        _, *content, __ = f.readlines()

    # Check the atom count is correct.
    num_atoms = len(self._atoms)
    if len(content) != num_atoms:
        raise RuntimeError(
            'The number of atoms in the coord file, '
            f'{len(content)}, does not match the number of atoms '
            f'in the molecule, {num_atoms}.'
        )

    # Save all the coords in the file.
    new_coords = []
    for i, line in enumerate(content):
        *coords, element = line.split()
        if element.isnumeric():
            element = periodic_table[int(element)]

        if element != self._atoms[i].__class__.__name__:
            raise RuntimeError(
                f'Atom {i} element does not match file.'
            )

        new_coords.append([float(i)*bohr_to_ang for i in coords])

    # Check that the correct number of atom
    # lines was present in the file.
    if i+1 != num_atoms:
        raise RuntimeError(
            'The number of atoms lines in the coord file, '
            f'{i+1}, does not match the number of atoms '
            f'in the molecule, {num_atoms}.'
        )

    # Update the structure.
    return self._with_position_matrix(new_coords)
